"""Linearity checking

Linearity checking across control-flow is done by the `CFGChecker`.
"""

import ast
from collections.abc import Generator, Iterator
from contextlib import contextmanager
from enum import Enum, auto
from typing import TYPE_CHECKING, NamedTuple, TypeGuard

from guppylang.ast_util import AstNode, find_nodes, get_type
from guppylang.cfg.analysis import LivenessAnalysis, LivenessDomain
from guppylang.cfg.bb import BB, VariableStats
from guppylang.checker.cfg_checker import CheckedBB, CheckedCFG, Row, Signature
from guppylang.checker.core import (
    FieldAccess,
    Globals,
    Locals,
    Place,
    PlaceId,
    SubscriptAccess,
    Variable,
    contains_subscript,
)
from guppylang.checker.errors.linearity import (
    AlreadyUsedError,
    BorrowShadowedError,
    BorrowSubPlaceUsedError,
    ComprAlreadyUsedError,
    DropAfterCallError,
    MoveOutOfSubscriptError,
    NonCopyableCaptureError,
    NonCopyablePartialApplyError,
    NotOwnedError,
    PlaceNotUsedError,
    UnnamedExprNotUsedError,
    UnnamedFieldNotUsedError,
    UnnamedSubscriptNotUsedError,
)
from guppylang.definition.custom import CustomFunctionDef
from guppylang.definition.value import CallableDef
from guppylang.error import GuppyError, GuppyTypeError
from guppylang.nodes import (
    AnyCall,
    BarrierExpr,
    CheckedNestedFunctionDef,
    DesugaredArrayComp,
    DesugaredGenerator,
    DesugaredListComp,
    FieldAccessAndDrop,
    GlobalCall,
    InoutReturnSentinel,
    LocalCall,
    PartialApply,
    PlaceNode,
    SubscriptAccessAndDrop,
    TensorCall,
)
from guppylang.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    StructType,
)

if TYPE_CHECKING:
    from guppylang.diagnostic import Error


class UseKind(Enum):
    """The different ways places can be used."""

    #: A classical value is copied
    COPY = auto()

    #: A value is borrowed when passing it to a function
    BORROW = auto()

    #: Ownership of an owned value is transferred by passing it to a function
    CONSUME = auto()

    #: Ownership of an owned value is transferred by returning it
    RETURN = auto()

    #: An owned value is renamed or stored in a tuple/list
    MOVE = auto()

    @property
    def indicative(self) -> str:
        """Describes a use in an indicative mood.

        For example: "You cannot *consume* this qubit."
        """
        return self.name.lower()

    @property
    def subjunctive(self) -> str:
        """Describes a use in a subjunctive mood.

        For example: "This qubit cannot be *consumed*"
        """
        match self:
            case UseKind.COPY:
                return "copied"
            case UseKind.BORROW:
                return "borrowed"
            case UseKind.CONSUME:
                return "consumed"
            case UseKind.RETURN:
                return "returned"
            case UseKind.MOVE:
                return "moved"


class Use(NamedTuple):
    """Records data associated with a use of a place."""

    #: The AST node corresponding to the use
    node: AstNode

    #: The kind of use, i.e. is the value consumed, borrowed, returned, ...?
    kind: UseKind


class Scope(Locals[PlaceId, Place]):
    """Scoped collection of assigned places indexed by their id.

    Keeps track of which places have already been used.
    """

    parent_scope: "Scope | None"
    used_local: dict[PlaceId, Use]
    used_parent: dict[PlaceId, Use]

    def __init__(self, parent: "Scope | None" = None):
        self.used_local = {}
        self.used_parent = {}
        super().__init__({}, parent)

    def used(self, x: PlaceId) -> Use | None:
        """Checks whether a place has already been used."""
        if x in self.vars:
            return self.used_local.get(x, None)
        assert self.parent_scope is not None
        return self.parent_scope.used(x)

    def use(self, x: PlaceId, node: AstNode, kind: UseKind) -> None:
        """Records a use of a place.

        Works for places in the current scope as well as places in any parent scope.
        """
        if x in self.vars:
            self.used_local[x] = Use(node, kind)
        else:
            assert self.parent_scope is not None
            assert x in self.parent_scope
            self.used_parent[x] = Use(node, kind)
            self.parent_scope.use(x, node, kind)

    def assign(self, place: Place) -> None:
        """Records an assignment of a place."""
        assert place.defined_at is not None
        x = place.id
        self.vars[x] = place
        if x in self.used_local:
            self.used_local.pop(x)

    def stats(self) -> VariableStats[PlaceId]:
        assigned = {}
        for x, place in self.vars.items():
            assert place.defined_at is not None
            assigned[x] = place.defined_at
        used = {x: use.node for x, use in self.used_parent.items()}
        return VariableStats(assigned, used)


class BBLinearityChecker(ast.NodeVisitor):
    """AST visitor that checks linearity for a single basic block."""

    scope: Scope
    stats: VariableStats[PlaceId]
    func_name: str
    func_inputs: dict[PlaceId, Variable]
    globals: Globals

    def check(
        self,
        bb: "CheckedBB[Variable]",
        is_entry: bool,
        func_name: str,
        func_inputs: dict[PlaceId, Variable],
        globals: Globals,
    ) -> Scope:
        # Manufacture a scope that holds all places that are live at the start
        # of this BB
        input_scope = Scope()
        for var in bb.sig.input_row:
            for place in leaf_places(var):
                input_scope.assign(place)
        self.func_name = func_name
        self.func_inputs = func_inputs
        self.globals = globals

        # Open up a new nested scope to check the BB contents. This way we can track
        # when we use variables from the outside vs ones assigned in this BB. The only
        # exception is the entry BB since function arguments should be treated as part
        # of the entry BB
        self.scope = input_scope if is_entry else Scope(input_scope)

        for stmt in bb.statements:
            self.visit(stmt)
        if bb.branch_pred:
            self.visit(bb.branch_pred)
        return self.scope

    @contextmanager
    def new_scope(self) -> Generator[Scope, None, None]:
        scope, new_scope = self.scope, Scope(self.scope)
        self.scope = new_scope
        yield new_scope
        self.scope = scope

    def visit_PlaceNode(
        self,
        node: PlaceNode,
        /,
        use_kind: UseKind = UseKind.MOVE,
        is_call_arg: AnyCall | None = None,
    ) -> None:
        # Usage of borrowed variables is generally forbidden. The only exception is
        # letting them be reborrowed by another function call. In that case, our
        # `_visit_call_args` helper will set `use_kind=UseKind.BORROW`.
        is_inout_arg = use_kind == UseKind.BORROW
        if is_inout_var(node.place) and not is_inout_arg:
            err: Error = NotOwnedError(
                node,
                node.place,
                use_kind,
                is_call_arg is not None,
                self._call_name(is_call_arg),
                self.func_name,
            )
            arg_span = self.func_inputs[node.place.root.id].defined_at
            err.add_sub_diagnostic(NotOwnedError.MakeOwned(arg_span))
            raise GuppyError(err)
        # Places involving subscripts are handled differently since we ignore everything
        # after the subscript for the purposes of linearity checking.
        if subscript := contains_subscript(node.place):
            # We have to check the item type to determine if we can move out of the
            # subscript.
            if not is_inout_arg and not subscript.ty.copyable:
                err = MoveOutOfSubscriptError(node, use_kind, subscript.parent)
                err.add_sub_diagnostic(MoveOutOfSubscriptError.Explanation(None))
                raise GuppyError(err)
            self.visit(subscript.item_expr)
            self.scope.assign(subscript.item)
            # Visiting the `__getitem__(place.parent, place.item)` call ensures that we
            # linearity-check the parent and element.
            assert subscript.getitem_call is not None
            self.visit(subscript.getitem_call)
        # For all other places, we record uses of all leaves
        else:
            for place in leaf_places(node.place):
                x = place.id
                if (prev_use := self.scope.used(x)) and not place.ty.copyable:
                    err = AlreadyUsedError(node, place, use_kind)
                    err.add_sub_diagnostic(
                        AlreadyUsedError.PrevUse(prev_use.node, prev_use.kind)
                    )
                    raise GuppyError(err)
                self.scope.use(x, node, use_kind)

    def visit_Assign(self, node: ast.Assign) -> None:
        self.visit(node.value)
        self._check_assign_targets(node.targets)

        # Check that borrowed vars are not being shadowed. This would also be caught by
        # the dataflow analysis later, however we can give nicer error messages here.
        [target] = node.targets
        for tgt in find_nodes(lambda n: isinstance(n, PlaceNode), target):
            assert isinstance(tgt, PlaceNode)
            if tgt.place.id in self.func_inputs:
                entry_place = self.func_inputs[tgt.place.id]
                if is_inout_var(entry_place):
                    err = BorrowShadowedError(tgt.place.defined_at, entry_place)
                    err.add_sub_diagnostic(BorrowShadowedError.Rename(None))
                    raise GuppyError(err)

    def visit_Return(self, node: ast.Return) -> None:
        # Intercept returns of places, so we can set the appropriate `use_kind` to get
        # nicer error messages
        if isinstance(node.value, PlaceNode):
            self.visit_PlaceNode(node.value, use_kind=UseKind.RETURN)
        elif isinstance(node.value, ast.Tuple):
            for elt in node.value.elts:
                if isinstance(elt, PlaceNode):
                    self.visit_PlaceNode(elt, use_kind=UseKind.RETURN)
                else:
                    self.visit(elt)
        elif node.value:
            self.visit(node.value)

    def _visit_call_args(self, func_ty: FunctionType, call: AnyCall) -> None:
        """Helper function to check the arguments of a function call.

        Populates the `use_kind` kwarg of `visit_PlaceNode` in case some of the
        arguments are places.
        """
        for inp, arg in zip(func_ty.inputs, call.args, strict=True):
            if isinstance(arg, PlaceNode):
                use_kind = (
                    UseKind.BORROW if InputFlags.Inout in inp.flags else UseKind.CONSUME
                )
                self.visit_PlaceNode(arg, use_kind=use_kind, is_call_arg=call)
            else:
                self.visit(arg)

    def _reassign_inout_args(self, func_ty: FunctionType, call: AnyCall) -> None:
        """Helper function to reassign the borrowed arguments after a function call."""
        for inp, arg in zip(func_ty.inputs, call.args, strict=True):
            if InputFlags.Inout in inp.flags:
                match arg:
                    case PlaceNode(place=place):
                        self._reassign_single_inout_arg(place, place.defined_at or arg)
                    case arg if not inp.ty.droppable:
                        err = DropAfterCallError(arg, inp.ty, self._call_name(call))
                        err.add_sub_diagnostic(DropAfterCallError.Assign(None))
                        raise GuppyError(err)

    def _reassign_single_inout_arg(self, place: Place, node: AstNode) -> None:
        """Helper function to reassign a single borrowed argument after a function
        call."""
        # Places involving subscripts are given back by visiting the `__setitem__` call
        if subscript := contains_subscript(place):
            assert subscript.setitem_call is not None
            for leaf in leaf_places(subscript.setitem_call.value_var):
                self.scope.assign(leaf)
            self.visit(subscript.setitem_call.call)
            self._reassign_single_inout_arg(subscript.parent, node)
        else:
            for leaf in leaf_places(place):
                assert not isinstance(leaf, SubscriptAccess)
                leaf = leaf.replace_defined_at(node)
                self.scope.assign(leaf)

    def _call_name(self, node: AnyCall | None) -> str | None:
        """Tries to extract the name of a called function from a call AST node."""
        if isinstance(node, LocalCall):
            return node.func.id if isinstance(node.func, ast.Name) else None
        elif isinstance(node, GlobalCall):
            return self.globals[node.def_id].name
        return None

    def visit_GlobalCall(self, node: GlobalCall) -> None:
        func = self.globals[node.def_id]
        assert isinstance(func, CallableDef)
        if isinstance(func, CustomFunctionDef) and not func.has_signature:
            func_ty = FunctionType(
                [FuncInput(get_type(arg), InputFlags.NoFlags) for arg in node.args],
                get_type(node),
            )
        else:
            func_ty = func.ty.instantiate(node.type_args)
        self._visit_call_args(func_ty, node)
        self._reassign_inout_args(func_ty, node)

    def visit_LocalCall(self, node: LocalCall) -> None:
        func_ty = get_type(node.func)
        assert isinstance(func_ty, FunctionType)
        self.visit(node.func)
        self._visit_call_args(func_ty, node)
        self._reassign_inout_args(func_ty, node)

    def visit_TensorCall(self, node: TensorCall) -> None:
        for arg in node.args:
            self.visit(arg)
        self._reassign_inout_args(node.tensor_ty, node)

    def visit_PartialApply(self, node: PartialApply) -> None:
        self.visit(node.func)
        for arg in node.args:
            ty = get_type(arg)
            if not ty.copyable:
                err = NonCopyablePartialApplyError(node)
                err.add_sub_diagnostic(NonCopyablePartialApplyError.Captured(arg, ty))
                raise GuppyError(err)
            self.visit(arg)

    def visit_FieldAccessAndDrop(self, node: FieldAccessAndDrop) -> None:
        # A field access on a value that is not a place. This means the value can no
        # longer be accessed after the field has been projected out. Thus, this is only
        # legal if there are no remaining linear fields on the value
        self.visit(node.value)
        for field in node.struct_ty.fields:
            if field.name != node.field.name and not field.ty.droppable:
                err = UnnamedFieldNotUsedError(node.value, field, node.struct_ty)
                err.add_sub_diagnostic(UnnamedFieldNotUsedError.Fix(None, node.field))
                raise GuppyError(err)

    def visit_SubscriptAccessAndDrop(self, node: SubscriptAccessAndDrop) -> None:
        # A subscript access on a value that is not a place. This means the value can no
        # longer be accessed after the item has been projected out. Thus, this is only
        # legal if the items in the container are not linear
        elem_ty = get_type(node.getitem_expr)
        if not elem_ty.droppable:
            value = node.original_expr.value
            err = UnnamedSubscriptNotUsedError(value, get_type(value))
            err.add_sub_diagnostic(
                UnnamedSubscriptNotUsedError.SubscriptHint(node.item_expr)
            )
            err.add_sub_diagnostic(UnnamedSubscriptNotUsedError.Fix(None))
            raise GuppyTypeError(err)
        self.visit(node.item_expr)
        self.scope.assign(node.item)
        self.visit(node.getitem_expr)

    def visit_BarrierExpr(self, node: BarrierExpr) -> None:
        self._visit_call_args(node.func_ty, node)
        self._reassign_inout_args(node.func_ty, node)

    def visit_Expr(self, node: ast.Expr) -> None:
        # An expression statement where the return value is discarded
        self.visit(node.value)
        ty = get_type(node.value)
        if not ty.droppable:
            err = UnnamedExprNotUsedError(node, ty)
            err.add_sub_diagnostic(UnnamedExprNotUsedError.Fix(None))
            raise GuppyTypeError(err)

    def visit_DesugaredListComp(self, node: DesugaredListComp) -> None:
        self._check_comprehension(node.generators, node.elt)

    def visit_DesugaredArrayComp(self, node: DesugaredArrayComp) -> None:
        self._check_comprehension([node.generator], node.elt)

    def visit_CheckedNestedFunctionDef(self, node: CheckedNestedFunctionDef) -> None:
        # Linearity of the nested function has already been checked. We just need to
        # verify that no linear variables are captured
        # TODO: In the future, we could support capturing of non-linear subplaces
        for var, use in node.captured.values():
            if not var.ty.copyable:
                err = NonCopyableCaptureError(use, var)
                err.add_sub_diagnostic(
                    NonCopyableCaptureError.DefinedHere(var.defined_at)
                )
                raise GuppyError(err)
            for place in leaf_places(var):
                self.scope.use(place.id, use, UseKind.COPY)
        self.scope.assign(Variable(node.name, node.ty, node))

    def _check_assign_targets(self, targets: list[ast.expr]) -> None:
        """Helper function to check assignments."""
        # We're not allowed to override an unused linear place
        [target] = targets
        for tgt in find_nodes(lambda n: isinstance(n, PlaceNode), target):
            assert isinstance(tgt, PlaceNode)
            # Special error message for shadowing of borrowed vars
            x = tgt.place.id
            if x in self.scope.vars and is_inout_var(self.scope[x]):
                err: Error = BorrowShadowedError(tgt, tgt.place)
                err.add_sub_diagnostic(BorrowShadowedError.Rename(None))
                raise GuppyError(err)
            # Subscript assignments also require checking the `__setitem__` call
            if subscript := contains_subscript(tgt.place):
                assert subscript.setitem_call is not None
                self.visit(subscript.item_expr)
                self.scope.assign(subscript.item)
                self.scope.assign(subscript.setitem_call.value_var)
                self.visit(subscript.setitem_call.call)
            else:
                for tgt_place in leaf_places(tgt.place):
                    x = tgt_place.id
                    # Only check for overrides of places locally defined in this BB.
                    # Global checks are handled by dataflow analysis.
                    if x in self.scope.vars and x not in self.scope.used_local:
                        place = self.scope[x]
                        if not place.ty.droppable:
                            err = PlaceNotUsedError(place.defined_at, place)
                            err.add_sub_diagnostic(PlaceNotUsedError.Fix(None))
                            raise GuppyError(err)
                    self.scope.assign(tgt_place)

    def _check_comprehension(
        self, gens: list[DesugaredGenerator], elt: ast.expr
    ) -> None:
        """Helper function to recursively check list comprehensions."""
        if not gens:
            self.visit(elt)
            return

        # Check the iterator expression in the current scope
        gen, *gens = gens
        self.visit(gen.iter_assign.value)
        assert isinstance(gen.iter, PlaceNode)

        # The rest is checked in a new nested scope so we can track which variables
        # are introduced and used inside the loop
        with self.new_scope() as inner_scope:
            # In particular, assign the iterator variable in the new scope
            self._check_assign_targets(gen.iter_assign.targets)
            self.visit(gen.next_call)
            self._check_assign_targets([gen.target])
            self._check_assign_targets(gen.iter_assign.targets)

            # `if` guards are generally not allowed when we're iterating over linear
            # variables. The only exception is if all linear variables are already
            # consumed by the first guard
            if gen.ifs:
                first_if, *other_ifs = gen.ifs
                # Check if there are linear iteration variables that have not been used
                # by the first guard
                self.visit(first_if)
                for place in self.scope.vars.values():
                    # The only exception is the iterator variable since we make sure
                    # that it is carried through each iteration during Hugr generation
                    if place == gen.iter.place:
                        continue
                    for leaf in leaf_places(place):
                        x = leaf.id
                        # Also ignore borrowed variables
                        if x in inner_scope.used_parent and (
                            inner_scope.used_parent[x].kind == UseKind.BORROW
                        ):
                            continue
                        if not self.scope.used(x) and not place.ty.droppable:
                            err = PlaceNotUsedError(place.defined_at, place)
                            err.add_sub_diagnostic(
                                PlaceNotUsedError.Branch(first_if, False)
                            )
                            raise GuppyTypeError(err)
                for expr in other_ifs:
                    self.visit(expr)

            # Recursively check the remaining generators
            self._check_comprehension(gens, elt)

            # Look for any linear variables that were borrowed from the outer scope
            gen.borrowed_outer_places = []
            for x, use in inner_scope.used_parent.items():
                if use.kind == UseKind.BORROW:
                    # Since `x` was borrowed, we know that is now also assigned in the
                    # inner scope since it gets reassigned in the local scope after the
                    # borrow expires
                    place = inner_scope[x]
                    gen.borrowed_outer_places.append(place)
                    # Also mark this place as implicitly used so we don't complain about
                    # it later.
                    for leaf in leaf_places(place):
                        inner_scope.use(
                            leaf.id, InoutReturnSentinel(leaf), UseKind.RETURN
                        )

            # Mark the iterator as used since it's carried into the next iteration
            for leaf in leaf_places(gen.iter.place):
                self.scope.use(leaf.id, gen.iter, UseKind.CONSUME)

            # We have to make sure that all linear variables that were introduced in the
            # inner scope have been used
            for place in inner_scope.vars.values():
                for leaf in leaf_places(place):
                    x = leaf.id
                    if not leaf.ty.droppable and not inner_scope.used(x):
                        raise GuppyTypeError(PlaceNotUsedError(leaf.defined_at, leaf))

        # On the other hand, we have to ensure that no linear places from the
        # outer scope have been used inside the comprehension (they would be used
        # multiple times since the comprehension body is executed repeatedly)
        for x, use in inner_scope.used_parent.items():
            place = inner_scope[x]
            # The only exception are values that are only borrowed from the outer
            # scope. These can be safely reassigned.
            if use.kind == UseKind.BORROW:
                self._reassign_single_inout_arg(place, use.node)
            elif not place.ty.copyable:
                raise GuppyTypeError(ComprAlreadyUsedError(use.node, place, use.kind))


def leaf_places(place: Place) -> Iterator[Place]:
    """Returns all leaf descendant projections of a place."""
    stack = [place]
    while stack:
        place = stack.pop()
        if isinstance(place.ty, StructType):
            stack += [
                FieldAccess(place, field, place.defined_at) for field in place.ty.fields
            ]
        else:
            yield place


def is_inout_var(place: Place) -> TypeGuard[Variable]:
    """Checks whether a place is a borrowed variable."""
    return isinstance(place, Variable) and InputFlags.Inout in place.flags


def check_cfg_linearity(
    cfg: "CheckedCFG[Variable]", func_name: str, globals: Globals
) -> "CheckedCFG[Place]":
    """Checks whether a CFG satisfies the linearity requirements.

    Raises a user-error if linearity violations are found.

    Returns a new CFG with refined basic block signatures in terms of *places* rather
    than just variables.
    """
    bb_checker = BBLinearityChecker()
    func_inputs: dict[PlaceId, Variable] = {v.id: v for v in cfg.entry_bb.sig.input_row}
    scopes: dict[BB, Scope] = {
        bb: bb_checker.check(
            bb,
            is_entry=bb == cfg.entry_bb,
            func_name=func_name,
            func_inputs=func_inputs,
            globals=globals,
        )
        for bb in cfg.bbs
    }

    # Mark the borrowed variables as implicitly used in the exit BB
    exit_scope = scopes[cfg.exit_bb]
    for var in cfg.entry_bb.sig.input_row:
        if InputFlags.Inout in var.flags:
            for leaf in leaf_places(var):
                exit_scope.use(leaf.id, InoutReturnSentinel(var=var), UseKind.RETURN)

    # Edge case: If the exit is unreachable, then the function will never terminate, so
    # there is no need to give the borrowed values back to the caller. To ensure that
    # the generated Hugr is still valid, we have to thread the borrowed arguments
    # through the non-terminating loop. We achieve this by considering borrowed
    # variables as live in every BB, even if the actual use in the exit is unreachable.
    # This is done by including borrowed vars in the initial value for the liveness
    # analysis below. The analogous thing was also done in the previous `CFG.analyze`
    # pass.
    live_default: LivenessDomain[PlaceId] = (
        {
            leaf.id: cfg.exit_bb
            for var in cfg.entry_bb.sig.input_row
            if InputFlags.Inout in var.flags
            for leaf in leaf_places(var)
        }
        if not cfg.exit_bb.reachable
        else {}
    )

    # Run liveness analysis with this initial value
    stats = {bb: scope.stats() for bb, scope in scopes.items()}
    live_before = LivenessAnalysis(
        stats, initial=live_default, include_unreachable=False
    ).run(cfg.bbs)

    # Construct a CFG that tracks places instead of just variables
    result_cfg: CheckedCFG[Place] = CheckedCFG(cfg.input_tys, cfg.output_ty)
    checked: dict[BB, CheckedBB[Place]] = {}

    for bb, scope in scopes.items():
        live_before_bb = live_before[bb]

        # We have to check that used not copyable variables are not being outputted
        for succ in bb.successors:
            live = live_before[succ]
            for x, use_bb in live.items():
                use_scope = scopes[use_bb]
                place = use_scope[x]
                if not place.ty.copyable and (prev_use := scope.used(x)):
                    use = use_scope.used_parent[x]
                    # Special case if this is a use arising from the implicit returning
                    # of a borrowed argument
                    if isinstance(use.node, InoutReturnSentinel):
                        assert isinstance(use.node.var, Variable)
                        assert InputFlags.Inout in use.node.var.flags
                        err: Error = BorrowSubPlaceUsedError(
                            use.node.var.defined_at, use.node.var, place
                        )
                        err.add_sub_diagnostic(
                            BorrowSubPlaceUsedError.PrevUse(
                                prev_use.node, prev_use.kind
                            )
                        )
                        err.add_sub_diagnostic(BorrowSubPlaceUsedError.Fix(None))
                        raise GuppyError(err)
                    err = AlreadyUsedError(use.node, place, use.kind)
                    err.add_sub_diagnostic(
                        AlreadyUsedError.PrevUse(prev_use.node, prev_use.kind)
                    )
                    raise GuppyError(err)

        # On the other hand, unused variables that are not droppable *must* be outputted
        for place in scope.values():
            for leaf in leaf_places(place):
                x = leaf.id
                # Some values are just in scope because the type checker determined
                # them as live in the first (less precises) dataflow analysis. It
                # might be the case that x is actually not live when considering
                # the second, more fine-grained, analysis based on places.
                if x not in live_before_bb and x not in scope.vars:
                    continue
                used_later = all(x in live_before[succ] for succ in bb.successors)
                if not leaf.ty.droppable and not scope.used(x) and not used_later:
                    err = PlaceNotUsedError(scope[x].defined_at, leaf)
                    # If there are some paths that lead to a consumption, we can give
                    # a nicer error message by highlighting the branch that leads to
                    # the leak
                    if any(x in live_before[succ] for succ in bb.successors):
                        assert bb.branch_pred is not None
                        [left_succ, _] = bb.successors
                        err.add_sub_diagnostic(
                            PlaceNotUsedError.Branch(
                                bb.branch_pred, x in live_before[left_succ]
                            )
                        )
                    err.add_sub_diagnostic(PlaceNotUsedError.Fix(None))
                    raise GuppyError(err)

        def live_places_row(bb: BB, original_row: Row[Variable]) -> Row[Place]:
            """Construct a row of all places that are live at the start of a given BB.

            The only exception are input and exit BBs whose signature should not be
            split up into places but instead keep the original variable signature.
            """
            if bb in (cfg.entry_bb, cfg.exit_bb):
                return original_row
            # N.B. we can ignore lint B023. The use of loop variable `scope` below is
            # safe since we don't call this function outside the loop.
            return [scope[x] for x in live_before[bb]]  # noqa: B023

        assert isinstance(bb, CheckedBB)
        sig = Signature(
            input_row=live_places_row(bb, bb.sig.input_row),
            output_rows=[
                live_places_row(succ, output_row)
                for succ, output_row in zip(
                    bb.successors, bb.sig.output_rows, strict=True
                )
            ],
        )
        checked[bb] = CheckedBB(
            bb.idx,
            result_cfg,
            bb.statements,
            branch_pred=bb.branch_pred,
            reachable=bb.reachable,
            sig=sig,
        )

    # Fill in missing fields of the result CFG
    result_cfg.bbs = list(checked.values())
    result_cfg.entry_bb = checked[cfg.entry_bb]
    result_cfg.exit_bb = checked[cfg.exit_bb]
    result_cfg.live_before = {checked[bb]: cfg.live_before[bb] for bb in cfg.bbs}
    result_cfg.ass_before = {checked[bb]: cfg.ass_before[bb] for bb in cfg.bbs}
    result_cfg.maybe_ass_before = {
        checked[bb]: cfg.maybe_ass_before[bb] for bb in cfg.bbs
    }
    for bb in cfg.bbs:
        checked[bb].predecessors = [checked[pred] for pred in bb.predecessors]
        checked[bb].successors = [checked[succ] for succ in bb.successors]
    return result_cfg



from typing import Generic, TypeVar

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")
T = guppy.type_var("T", module=module)


@guppy.struct(module)
class MyStruct(Generic[T]):
    x: list[T]


@guppy(module)
def foo(s: MyStruct) -> None:
    pass


module.compile()



from typing import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum_functional import h

import guppylang.std.quantum as quantum

module = GuppyModule("test")
module.load_all(quantum)
module.load(h)

T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo(x: Callable[[T], T]) -> None:
    ...


@guppy(module)
def main() -> None:
    foo(h)


module.compile()



import ast
import inspect
from collections.abc import Sequence
from dataclasses import dataclass, field, replace
from typing import Any, ClassVar

from hugr import Wire, ops

from guppylang.ast_util import AstNode, annotate_location
from guppylang.checker.core import Globals, PyScope
from guppylang.checker.errors.generic import (
    ExpectedError,
    UnexpectedError,
    UnsupportedError,
)
from guppylang.definition.common import (
    CheckableDef,
    CompiledDef,
    DefId,
    Definition,
    ParsableDef,
    UnknownSourceError,
)
from guppylang.definition.custom import (
    CustomCallCompiler,
    CustomFunctionDef,
    DefaultCallChecker,
)
from guppylang.definition.function import parse_source
from guppylang.definition.parameter import ParamDef
from guppylang.definition.ty import TypeDef
from guppylang.diagnostic import Error, Help
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.ipython_inspect import find_ipython_def, is_running_ipython
from guppylang.span import SourceMap
from guppylang.tys.arg import Argument
from guppylang.tys.param import Parameter, check_all_args
from guppylang.tys.parsing import type_from_ast
from guppylang.tys.ty import FuncInput, FunctionType, InputFlags, StructType, Type


@dataclass(frozen=True)
class UncheckedStructField:
    """A single field on a struct whose type has not been checked yet."""

    name: str
    type_ast: ast.expr


@dataclass(frozen=True)
class StructField:
    """A single field on a struct."""

    name: str
    ty: Type


@dataclass(frozen=True)
class DuplicateFieldError(Error):
    title: ClassVar[str] = "Duplicate field"
    span_label: ClassVar[str] = (
        "Struct `{struct_name}` already contains a field named `{field_name}`"
    )
    struct_name: str
    field_name: str


@dataclass(frozen=True)
class NonGuppyMethodError(Error):
    title: ClassVar[str] = "Not a Guppy method"
    span_label: ClassVar[str] = (
        "Method `{method_name}` of struct `{struct_name}` is not a Guppy function"
    )
    struct_name: str
    method_name: str

    @dataclass(frozen=True)
    class Suggestion(Help):
        message: ClassVar[str] = (
            "Add a `@guppy` annotation to turn `{method_name}` into a Guppy method"
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(NonGuppyMethodError.Suggestion(None))


@dataclass(frozen=True)
class RawStructDef(TypeDef, ParsableDef):
    """A raw struct type definition that has not been parsed yet."""

    python_class: type
    python_scope: PyScope = field(repr=False)

    def __getitem__(self, item: Any) -> "RawStructDef":
        """Dummy implementation to enable subscripting in the Python runtime.

        For example, if users write `MyStruct[int]` in a function signature, the
        interpreter will try to execute the expression which would fail if this function
        weren't implemented.
        """
        return self

    def parse(self, globals: Globals, sources: SourceMap) -> "ParsedStructDef":
        """Parses the raw class object into an AST and checks that it is well-formed."""
        cls_def = parse_py_class(self.python_class, sources)
        if cls_def.keywords:
            raise GuppyError(UnexpectedError(cls_def.keywords[0], "keyword"))

        # The only base we allow is `Generic[...]` to specify generic parameters
        # TODO: This will become obsolete once we have Python 3.12 style generic classes
        params: list[Parameter]
        match cls_def.bases:
            case []:
                params = []
            case [base] if elems := try_parse_generic_base(base):
                params = params_from_ast(elems, globals)
            case bases:
                err: Error = UnsupportedError(
                    bases[0], "Struct inheritance", singular=True
                )
                raise GuppyError(err)

        fields: list[UncheckedStructField] = []
        used_field_names: set[str] = set()
        used_func_names: dict[str, ast.FunctionDef] = {}
        for i, node in enumerate(cls_def.body):
            match i, node:
                # We allow `pass` statements to define empty structs
                case _, ast.Pass():
                    pass
                # Docstrings are also fine if they occur at the start
                case 0, ast.Expr(value=ast.Constant(value=v)) if isinstance(v, str):
                    pass
                # Ensure that all function definitions are Guppy functions
                case _, ast.FunctionDef(name=name) as node:
                    v = getattr(self.python_class, name)
                    if not isinstance(v, Definition):
                        raise GuppyError(NonGuppyMethodError(node, self.name, name))
                    used_func_names[name] = node
                    if name in used_field_names:
                        raise GuppyError(DuplicateFieldError(node, self.name, name))
                # Struct fields are declared via annotated assignments without value
                case _, ast.AnnAssign(target=ast.Name(id=field_name)) as node:
                    if node.value:
                        err = UnsupportedError(node.value, "Default struct values")
                        raise GuppyError(err)
                    if field_name in used_field_names:
                        err = DuplicateFieldError(node.target, self.name, field_name)
                        raise GuppyError(err)
                    fields.append(UncheckedStructField(field_name, node.annotation))
                    used_field_names.add(field_name)
                case _, node:
                    err = UnexpectedError(
                        node, "statement", unexpected_in="struct definition"
                    )
                    raise GuppyError(err)

        # Ensure that functions don't override struct fields
        if overridden := used_field_names.intersection(used_func_names.keys()):
            x = overridden.pop()
            raise GuppyError(DuplicateFieldError(used_func_names[x], self.name, x))

        return ParsedStructDef(
            self.id, self.name, cls_def, params, fields, self.python_scope
        )

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> Type:
        raise InternalGuppyError("Tried to instantiate raw struct definition")


@dataclass(frozen=True)
class ParsedStructDef(TypeDef, CheckableDef):
    """A struct definition whose fields have not been checked yet."""

    defined_at: ast.ClassDef
    params: Sequence[Parameter]
    fields: Sequence[UncheckedStructField]
    python_scope: PyScope = field(repr=False)

    def check(self, globals: Globals) -> "CheckedStructDef":
        """Checks that all struct fields have valid types."""
        # Before checking the fields, make sure that this definition is not recursive,
        # otherwise the code below would not terminate.
        # TODO: This is not ideal (see todo in `check_instantiate`)
        globals = globals.with_python_scope(self.python_scope)
        param_var_mapping = {p.name: p for p in self.params}
        check_not_recursive(self, globals, param_var_mapping)

        fields = [
            StructField(f.name, type_from_ast(f.type_ast, globals, param_var_mapping))
            for f in self.fields
        ]
        return CheckedStructDef(
            self.id, self.name, self.defined_at, self.params, fields
        )

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> Type:
        """Checks if the struct can be instantiated with the given arguments."""
        check_all_args(self.params, args, self.name, loc)
        # Obtain a checked version of this struct definition so we can construct a
        # `StructType` instance
        # TODO: This is quite bad: If we have a cyclic definition this will not
        #  terminate, so we have to check for cycles in every call to `check`. The
        #  proper way to deal with this is changing `StructType` such that it only
        #  takes a `DefId` instead of a `CheckedStructDef`. But this will be a bigger
        #  refactor...
        checked_def = self.check(globals)
        return StructType(args, checked_def)


@dataclass(frozen=True)
class CheckedStructDef(TypeDef, CompiledDef):
    """A struct definition that has been fully checked."""

    defined_at: ast.ClassDef
    params: Sequence[Parameter]
    fields: Sequence[StructField]

    def check_instantiate(
        self, args: Sequence[Argument], globals: "Globals", loc: AstNode | None = None
    ) -> Type:
        """Checks if the struct can be instantiated with the given arguments."""
        check_all_args(self.params, args, self.name, loc)
        return StructType(args, self)

    def generated_methods(self) -> list[CustomFunctionDef]:
        """Auto-generated methods for this struct."""

        class ConstructorCompiler(CustomCallCompiler):
            """Compiler for the `__new__` constructor method of a struct."""

            def compile(self, args: list[Wire]) -> list[Wire]:
                return list(self.builder.add(ops.MakeTuple()(*args)))

        constructor_sig = FunctionType(
            inputs=[
                FuncInput(f.ty, InputFlags.Owned if f.ty.linear else InputFlags.NoFlags)
                for f in self.fields
            ],
            output=StructType(
                defn=self, args=[p.to_bound(i) for i, p in enumerate(self.params)]
            ),
            input_names=[f.name for f in self.fields],
            params=self.params,
        )
        constructor_def = CustomFunctionDef(
            id=DefId.fresh(self.id.module),
            name="__new__",
            defined_at=self.defined_at,
            ty=constructor_sig,
            call_checker=DefaultCallChecker(),
            call_compiler=ConstructorCompiler(),
            higher_order_value=True,
            has_signature=True,
        )
        return [constructor_def]


def parse_py_class(cls: type, sources: SourceMap) -> ast.ClassDef:
    """Parses a Python class object into an AST."""
    # If we are running IPython, `inspect.getsourcelines` works only for builtins
    # (guppy stdlib), but not for most/user-defined classes - see:
    #  - https://bugs.python.org/issue33826
    #  - https://github.com/ipython/ipython/issues/11249
    #  - https://github.com/wandb/weave/pull/1864
    if is_running_ipython():
        defn = find_ipython_def(cls.__name__)
        if defn is not None:
            file_name = f"<{defn.cell_name}>"
            sources.add_file(file_name, defn.cell_source)
            annotate_location(defn.node, defn.cell_source, file_name, 1)
            if not isinstance(defn.node, ast.ClassDef):
                raise GuppyError(ExpectedError(defn.node, "a class definition"))
            return defn.node
        # else, fall through to handle builtins.
    source_lines, line_offset = inspect.getsourcelines(cls)
    source, cls_ast, line_offset = parse_source(source_lines, line_offset)
    file = inspect.getsourcefile(cls)
    if file is None:
        raise GuppyError(UnknownSourceError(None, cls))
    # Store the source file in our cache
    sources.add_file(file)
    annotate_location(cls_ast, source, file, line_offset)
    if not isinstance(cls_ast, ast.ClassDef):
        raise GuppyError(ExpectedError(cls_ast, "a class definition"))
    return cls_ast


def try_parse_generic_base(node: ast.expr) -> list[ast.expr] | None:
    """Checks if an AST node corresponds to a `Generic[T1, ..., Tn]` base class.

    Returns the generic parameters or `None` if the AST has a different shape
    """
    match node:
        case ast.Subscript(value=ast.Name(id="Generic"), slice=elem):
            return elem.elts if isinstance(elem, ast.Tuple) else [elem]
        case _:
            return None


@dataclass(frozen=True)
class RepeatedTypeParamError(Error):
    title: ClassVar[str] = "Duplicate type parameter"
    span_label: ClassVar[str] = "Type parameter `{name}` cannot be used multiple times"
    name: str


def params_from_ast(nodes: Sequence[ast.expr], globals: Globals) -> list[Parameter]:
    """Parses a list of AST nodes into unique type parameters.

    Raises user errors if the AST nodes don't correspond to parameters or parameters
    occur multiple times.
    """
    params: list[Parameter] = []
    params_set: set[DefId] = set()
    for node in nodes:
        if isinstance(node, ast.Name) and node.id in globals:
            defn = globals[node.id]
            if isinstance(defn, ParamDef):
                if defn.id in params_set:
                    raise GuppyError(RepeatedTypeParamError(node, node.id))
                params.append(defn.to_param(len(params)))
                params_set.add(defn.id)
                continue
        raise GuppyError(ExpectedError(node, "a type parameter"))
    return params


def check_not_recursive(
    defn: ParsedStructDef, globals: Globals, param_var_mapping: dict[str, Parameter]
) -> None:
    """Throws a user error if the given struct definition is recursive."""

    # TODO: The implementation below hijacks the type parsing logic to detect recursive
    #  structs. This is not great since it repeats the work done during checking. We can
    #  get rid of this after resolving the todo in `ParsedStructDef.check_instantiate()`

    @dataclass(frozen=True)
    class DummyStructDef(TypeDef):
        """Dummy definition that throws an error when trying to instantiate it.

        By replacing the defn with this, we can detect recursive occurrences during
        type parsing.
        """

        def check_instantiate(
            self,
            args: Sequence[Argument],
            globals: "Globals",
            loc: AstNode | None = None,
        ) -> Type:
            raise GuppyError(UnsupportedError(loc, "Recursive structs"))

    dummy_defs = {
        **globals.defs,
        defn.id: DummyStructDef(defn.id, defn.name, defn.defined_at),
    }
    dummy_globals = replace(globals, defs=globals.defs | dummy_defs)
    for fld in defn.fields:
        type_from_ast(fld.type_ast, dummy_globals, param_var_mapping)



import tests.error.util
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from tests.error.util import NonBool

module = GuppyModule("test")
module.load_all(tests.error.util)


@guppy(module)
def foo(x: NonBool) -> int:
    while x:
        pass
    return 0


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")


@guppy.struct(module)
class MyStruct:
    x: int
    """Docstring in wrong position"""
    y: bool


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array, owned


module = GuppyModule("test")

@guppy(module)
def main(xs: array[int, 2] @ owned) -> array[int, 2]:
   ys = xs
   return xs

module.compile()



"""
Implementation of the adaptive random walk phase estimation algorithm from
https://arxiv.org/abs/2208.04526.

The example Hamiltonian and numbers are taken from https://arxiv.org/abs/2206.12950.
"""

import math
from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.std.angles import angle
from guppylang.std.builtins import owned, py, result
from guppylang.std.quantum import discard, measure, qubit
from guppylang.std.quantum_functional import cx, h, rz, x

sqrt_e = math.sqrt(math.e)
sqrt_e_div = math.sqrt((math.e - 1) / math.e)


@guppy
def random_walk_phase_estimation(
    eigenstate: Callable[[], qubit],
    controlled_oracle: Callable[[qubit @owned, qubit @owned, float], tuple[qubit, qubit]],
    num_iters: int,
    reset_rate: int,
    mu: float,
    sigma: float,
) -> float:
    """Performs the random walk phase estimation algorithm on a single qubit for
    some Hamiltonian H.

    Arguments:
       eigenstate: Function preparing the eigenstate of e^itH
       controlled_oracle: The oracle circuit for a controlled e^itH
       num_iters: Number of iterations to run the algorithm for
       reset_rate: Reset the eigenstate every x iterations
       mu: Initial mean for the eigenvalue estimate
       sigma: Initial standard deviation for the eigenvalue estimate
    """
    tgt = eigenstate()
    i = 0
    while i < num_iters:
        aux = h(qubit())
        t = 1 / sigma
        aux = rz(h(aux), angle((sigma - mu) * t))
        aux, tgt = controlled_oracle(aux, tgt, t)
        if measure(h(aux)):
            mu += sigma / py(sqrt_e)
        else:
            mu -= sigma / py(sqrt_e)
        sigma *= py(sqrt_e_div)

        # Reset the eigenstate every few iterations to increase the fidelity of
        # the algorithm
        if i % reset_rate == 0:
            discard(tgt)
            tgt = eigenstate()
        i += 1
    discard(tgt)
    return mu


@guppy
def example_controlled_oracle(q1: qubit @owned, q2: qubit @owned, t: float) -> tuple[qubit, qubit]:
    """A controlled e^itH gate for the example Hamiltonian H = -0.5 * Z"""
    # This is just a controlled rz gate
    a = angle(-0.5 * t)
    q2 = rz(q2, a / 2)
    q1, q2 = cx(q1, q2)
    q2 = rz(q2, -a / 2)
    return cx(q1, q2)


@guppy
def example_eigenstate() -> qubit:
    """The eigenstate of e^itH for the example Hamiltonian H = -0.5 * Z"""
    # This is just |1>
    return x(qubit())


@guppy
def main() -> int:
    num_iters = 24  # To avoid underflows
    reset_rate = 8
    mu = py(sqrt_e)
    sigma = py(sqrt_e_div)
    eigenvalue = random_walk_phase_estimation(
        example_eigenstate,
        example_controlled_oracle,
        num_iters,
        reset_rate,
        mu,
        sigma,
    )
    result("eigenvalue", eigenvalue)  # Expected outcome is 0.5
    return 0


hugr = guppy.compile_module()



from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def foo(xs: list[int]) -> int:
        for x in xs:
            pass
        return 0

    validate(foo)


def test_counting_loop(validate):
    @compile_guppy
    def foo(xs: list[int]) -> int:
        s = 0
        for x in xs:
            s += x
        return s

    validate(foo)


def test_multi_targets(validate):
    @compile_guppy
    def foo(xs: list[tuple[int, float]]) -> float:
        s = 0.0
        for x, y in xs:
            s += x * y
        return s

    validate(foo)


def test_multi_targets_same(validate):
    @compile_guppy
    def foo(xs: list[tuple[int, float]]) -> float:
        s = 1.0
        for x, x in xs:
            s *= x
        return s

    validate(foo)


def test_reassign_iter(validate):
    @compile_guppy
    def foo(xs: list[int]) -> int:
        s = 1
        for x in xs:
            xs = False
            s += x
        return s

    validate(foo)


def test_break(validate):
    @compile_guppy
    def foo(xs: list[int]) -> int:
        i = 1
        for x in xs:
            if x >= 42:
                break
            i *= x
        return i

    validate(foo)


def test_continue(validate):
    @compile_guppy
    def foo(xs: list[int]) -> int:
        i = len(xs)
        for x in xs:
            if x >= 42:
                continue
            i -= 1
        return i

    validate(foo)


def test_return_in_loop(validate):
    @compile_guppy
    def foo(xs: list[int]) -> int:
        y = 42
        for x in xs:
            if x >= 1337:
                return x * y
            y = y + x
        return y

    validate(foo)


def test_nested_loop(validate):
    @compile_guppy
    def foo(xs: list[int], ys: list[int]) -> int:
        p = 0
        for x in xs:
            s = 0
            for y in ys:
                s += y * x
            p += s - x
        return p

    validate(foo)


def test_nested_loop_break_continue(validate):
    @compile_guppy
    def foo(xs: list[int], ys: list[int]) -> int:
        p = 0
        for x in xs:
            s = 0
            for y in ys:
                if x % 2 == 0:
                    continue
                s += x
                if s > y:
                    s = y
                else:
                    break
            p += s * x
        return p

    validate(foo)



from guppylang import GuppyModule, guppy
from guppylang.std.builtins import panic
from tests.util import compile_guppy


def test_basic(validate):
    @compile_guppy
    def main() -> None:
        panic("I panicked!")

    validate(main)


def test_discard(validate):
    @compile_guppy
    def main() -> None:
        a = 1 + 2
        panic("I panicked!", False, a)

    validate(main)


def test_value(validate):
    module = GuppyModule("test")

    @guppy(module)
    def foo() -> int:
        return panic("I panicked!")

    @guppy(module)
    def bar() -> tuple[int, float]:
        return panic("I panicked!")

    @guppy(module)
    def baz() -> None:
        return panic("I panicked!")

    validate(module.compile())

def test_py_message(validate):
    @compile_guppy
    def main(x: int) -> None:
        panic(py("I" + "panicked" + "!"))

    validate(main)



import pytest

from guppylang import GuppyModule, guppy


@pytest.mark.xfail
def test_hand(validate):
    """Regression test ensuring that pytest terminates, even if the Guppy compiler
    throws an error.

    See https://github.com/CQCL/guppylang/issues/569
    """
    module = GuppyModule("test")

    @guppy(module)
    def test() -> int:
        return a  # Intentional use of an undefined variable

    validate(module.compile())



from guppylang import GuppyModule, guppy
from guppylang.std.builtins import array


module = GuppyModule('main')
n = guppy.nat_var("n", module=module)


@guppy(module)
def foo(xs: array[int, n]) -> int:
    a, *bs = xs
    return a


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def bar(q: qubit) -> bool:
    ...


@guppy(module)
def foo(qs: list[tuple[bool, qubit]] @owned) -> list[int]:
    return [42 for b, q in qs if b if bar(q)]


module.compile()



"""Utilities for defining builtin functions.

Note: These custom definitions will be replaced with direct extension operation
definitions from the hugr library.
"""

from collections.abc import Callable
from typing import TYPE_CHECKING

import hugr.std.collections.list
import hugr.std.float
import hugr.std.int
import hugr.std.logic
from hugr import ext as he
from hugr import ops
from hugr import tys as ht

from guppylang.compiler.hugr_extension import UnsupportedOp
from guppylang.std._internal.compiler.quantum import QUANTUM_EXTENSION
from guppylang.tys.subst import Inst
from guppylang.tys.ty import NumericType

if TYPE_CHECKING:
    from guppylang.tys.arg import Argument


def int_arg(n: int = NumericType.INT_WIDTH) -> ht.TypeArg:
    """A bounded int type argument."""
    return ht.BoundedNatArg(n=n)


def type_arg(idx: int = 0, bound: ht.TypeBound = ht.TypeBound.Any) -> ht.TypeArg:
    """A generic type argument."""
    return ht.VariableArg(idx=idx, param=ht.TypeTypeParam(bound=bound))


def make_concrete_arg(
    arg: ht.TypeArg,
    inst: Inst,
    variable_remap: dict[int, int] | None = None,
) -> ht.TypeArg:
    """Makes a concrete hugr type argument using a guppy instantiation.

    Args:
        arg: The hugr type argument to make concrete, containing variable arguments.
        inst: The guppy instantiation of the type arguments.
        variable_remap: A mapping from the hugr param variable indices to
            de Bruijn indices in the guppy type. Defaults to identity.
    """
    remap = variable_remap or {}

    if isinstance(arg, ht.VariableArg) and remap.get(arg.idx, arg.idx) < len(inst):
        concrete_arg: Argument = inst[remap.get(arg.idx, arg.idx)]
        return concrete_arg.to_hugr()
    return arg


def external_op(
    name: str,
    args: list[ht.TypeArg],
    ext: he.Extension,
    *,
    variable_remap: dict[int, int] | None = None,
) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Custom hugr operation

    Args:
        op_name: The name of the operation.
        args: The type arguments of the operation.
        ext: The extension of the operation.
        variable_remap: A mapping from the hugr param variable indices to
            de Bruijn indices in the guppy type. Defaults to identity.

    Returns:
        A function that takes an instantiation of the type arguments as well as
        the inferred input and output types and returns a concrete HUGR op.
    """
    op_def = ext.get_op(name)

    def op(ty: ht.FunctionType, inst: Inst) -> ops.DataflowOp:
        concrete_args = [make_concrete_arg(arg, inst, variable_remap) for arg in args]
        return op_def.instantiate(concrete_args, ty)

    return op


def float_op(
    op_name: str,
    ext: he.Extension = hugr.std.float.FLOAT_OPS_EXTENSION,
) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Utility method to create Hugr float arithmetic ops.

    Args:
        op_name: The name of the operation.
        ext: The extension of the operation.

    Returns:
        A function that takes an instantiation of the type arguments and returns
        a concrete HUGR op.
    """
    return external_op(op_name, args=[], ext=ext, variable_remap=None)


def int_op(
    op_name: str,
    ext: he.Extension = hugr.std.int.INT_OPS_EXTENSION,
    n_vars: int = 1,
) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Utility method to create Hugr integer arithmetic ops.

    Args:
        op_name: The name of the operation.
        ext: The extension of the operation.
        n_vars: The number of type arguments. Defaults to 1.

    Returns:
        A function that takes an instantiation of the type arguments and returns
        a concrete HUGR op.
    """
    # Ideally we'd be able to derive the arguments from the input/output types,
    # but the amount of variables does not correlate with the signature for the
    # integer ops in hugr :/
    # https://github.com/CQCL/hugr/blob/bfa13e59468feb0fc746677ea3b3a4341b2ed42e/hugr-core/src/std_extensions/arithmetic/int_ops.rs#L116
    #
    # For now, we just instantiate every type argument to a 64-bit integer.
    args: list[ht.TypeArg] = [int_arg() for _ in range(n_vars)]

    return external_op(
        op_name,
        args=args,
        ext=ext,
        variable_remap=None,
    )


def logic_op(
    op_name: str,
    parametric_size: bool = False,
    ext: he.Extension = hugr.std.logic.EXTENSION,
) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Utility method to create Hugr logic ops.

    If `parametric_size` is True, the generated operations has a single argument
    encoding the number of boolean inputs to the operation.

    args:
        op_name: The name of the operation.
        parametric_size: Whether the input count is a parameter to the operation.
        ext: The extension of the operation.

    Returns:
        A function that takes an instantiation of the type arguments and returns
        a concrete HUGR op.
    """
    op_def = ext.get_op(op_name)

    def op(ty: ht.FunctionType, inst: Inst) -> ops.DataflowOp:
        args = [int_arg(len(ty.input))] if parametric_size else []
        return ops.ExtOp(
            op_def,
            ty,
            args,
        )

    return op


def list_op(
    op_name: str,
    ext: he.Extension = hugr.std.collections.list.EXTENSION,
) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Utility method to create Hugr list ops.

    Args:
        op_name: The name of the operation.
        ext: The extension of the operation.

    Returns:
        A function that takes an instantiation of the type arguments and returns
        a concrete HUGR op.
    """
    op_def = ext.get_op(op_name)

    def op(ty: ht.FunctionType, inst: Inst) -> ops.DataflowOp:
        return ops.ExtOp(
            op_def,
            ty,
            [arg.to_hugr() for arg in inst],
        )

    return op


def quantum_op(
    op_name: str,
    ext: he.Extension = QUANTUM_EXTENSION,
) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Utility method to create Hugr quantum ops.

    args:
        op_name: The name of the operation.
        ext: The extension of the operation.

    Returns:
        A function that takes an instantiation of the type arguments and returns
        a concrete HUGR op.
    """
    op_def = ext.get_op(op_name)

    def op(ty: ht.FunctionType, inst: Inst) -> ops.DataflowOp:
        return ops.ExtOp(
            op_def,
            ty,
            args=[],
        )

    return op


def unsupported_op(op_name: str) -> Callable[[ht.FunctionType, Inst], ops.DataflowOp]:
    """Utility method to define not-yet-implemented operations.

    Args:
        op_name: The name of the operation to define.

    Returns:
        A function that takes an instantiation of the type arguments as well as
        the inferred input and output types and returns a concrete HUGR op.
    """

    def op(ty: ht.FunctionType, inst: Inst) -> ops.DataflowOp:
        return UnsupportedOp(
            op_name=op_name,
            inputs=ty.input,
            outputs=ty.output,
        )

    return op



from guppylang.decorator import guppy
from guppylang.module import GuppyModule

mod = GuppyModule("test")

x = guppy.extern("x", ty="int", module=mod)

@guppy(mod)
def bad(b: bool) -> int:
    if b:
        x = 4
    return x

mod.compile()



from guppylang import GuppyModule, guppy, qubit
from guppylang.std import quantum
from guppylang.std.quantum import discard
from tests.util import compile_guppy


def test_var_defined1(validate):
    @compile_guppy
    def test() -> int:
        if True:
            x = 1
        return x

    validate(test)


def test_var_defined2(validate):
    @compile_guppy
    def test(b: bool) -> int:
        while True:
            if b:
                x = 1
                break
        return x

    validate(test)


def test_type_mismatch1(validate):
    @compile_guppy
    def test() -> int:
        if True:
            x = 1
        else:
            x = 1.0
        return x

    validate(test)


def test_type_mismatch2(validate):
    @compile_guppy
    def test() -> int:
        x = 1
        while False:
            x = 1.0
        return x

    validate(test)


def test_type_mismatch3(validate):
    @compile_guppy
    def test() -> int:
        x = 1
        if False and (x := 1.0):
            pass
        return x

    validate(test)


def test_unused_var_use1(validate):
    @compile_guppy
    def test() -> int:
        x = 1
        if True:
            return 0
        return x

    validate(test)


def test_unused_var_use2(validate):
    @compile_guppy
    def test() -> int:
        x = 1
        if not False:
            x = 1.0
            return 0
        return x

    validate(test)


def test_unreachable_leak(validate):
    module = GuppyModule("module")
    module.load_all(quantum)

    @guppy(module)
    def test(b: bool) -> int:
        q = qubit()
        while True:
            if b:
                discard(q)
                return 1
        # This return would leak, but we don't complain since it's unreachable:
        return 0

    validate(module.compile())


def test_unreachable_leak2(validate):
    module = GuppyModule("module")
    module.load_all(quantum)

    @guppy(module)
    def test() -> None:
        if False:
            # This would leak, but we don't complain since it's unreachable:
            q = qubit()

    validate(module.compile())


def test_unreachable_copy(validate):
    module = GuppyModule("module")
    module.load_all(quantum)

    @guppy(module)
    def test() -> None:
        q = qubit()
        discard(q)
        if False:
            # This would be a linearity violation, but we don't complain since it's
            # unreachable:
            h(q)

    validate(module.compile())



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import array

module = GuppyModule("test")

S = guppy.type_var("S", module=module)
T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo(x: S, y: T) -> None:
    ...


@guppy(module)
def main(x: int, y: float) -> None:
    foo[float, int](x, y)


module.compile()



"""Type checking code for statements.

Operates on statements in a basic block after CFG construction. In particular, we
assume that statements involving control flow (i.e. if, while, break, and return
statements) have been removed during CFG construction.

After checking, we return a desugared statement where all sub-expression have been type
annotated.
"""

import ast
import functools
from collections.abc import Iterable, Sequence
from dataclasses import replace
from itertools import takewhile
from typing import TypeVar, cast

from guppylang.ast_util import (
    AstVisitor,
    get_type,
    with_loc,
    with_type,
)
from guppylang.cfg.bb import BB, BBStatement
from guppylang.cfg.builder import (
    desugar_comprehension,
    make_var,
    tmp_vars,
)
from guppylang.checker.core import (
    Context,
    FieldAccess,
    SubscriptAccess,
    Variable,
)
from guppylang.checker.errors.generic import UnsupportedError
from guppylang.checker.errors.type_errors import (
    AssignFieldTypeMismatchError,
    AssignNonPlaceHelp,
    AssignSubscriptTypeMismatchError,
    AttributeNotFoundError,
    MissingReturnValueError,
    StarredTupleUnpackError,
    TypeInferenceError,
    UnpackableError,
    WrongNumberOfUnpacksError,
)
from guppylang.checker.expr_checker import (
    ExprChecker,
    ExprSynthesizer,
    check_place_assignable,
    synthesize_comprehension,
)
from guppylang.error import GuppyError, GuppyTypeError, InternalGuppyError
from guppylang.nodes import (
    AnyUnpack,
    DesugaredArrayComp,
    IterableUnpack,
    MakeIter,
    NestedFunctionDef,
    PlaceNode,
    TupleUnpack,
    UnpackPattern,
)
from guppylang.span import Span, to_span
from guppylang.tys.builtin import (
    array_type,
    get_element_type,
    get_iter_size,
    is_array_type,
    is_sized_iter_type,
    nat_type,
)
from guppylang.tys.const import ConstValue
from guppylang.tys.parsing import type_from_ast
from guppylang.tys.subst import Subst
from guppylang.tys.ty import (
    ExistentialTypeVar,
    FunctionType,
    NoneType,
    StructType,
    TupleType,
    Type,
)


class StmtChecker(AstVisitor[BBStatement]):
    ctx: Context
    bb: BB | None
    return_ty: Type | None

    def __init__(
        self, ctx: Context, bb: BB | None = None, return_ty: Type | None = None
    ) -> None:
        assert not return_ty or not return_ty.unsolved_vars
        self.ctx = ctx
        self.bb = bb
        self.return_ty = return_ty

    def check_stmts(self, stmts: Sequence[BBStatement]) -> list[BBStatement]:
        return [self.visit(s) for s in stmts]

    def _synth_expr(self, node: ast.expr) -> tuple[ast.expr, Type]:
        return ExprSynthesizer(self.ctx).synthesize(node)

    def _synth_instance_fun(
        self,
        node: ast.expr,
        args: list[ast.expr],
        func_name: str,
        description: str,
        exp_sig: FunctionType | None = None,
        give_reason: bool = False,
    ) -> tuple[ast.expr, Type]:
        return ExprSynthesizer(self.ctx).synthesize_instance_func(
            node, args, func_name, description, exp_sig, give_reason
        )

    def _check_expr(
        self, node: ast.expr, ty: Type, kind: str = "expression"
    ) -> tuple[ast.expr, Subst]:
        return ExprChecker(self.ctx).check(node, ty, kind)

    @functools.singledispatchmethod
    def _check_assign(self, lhs: ast.expr, rhs: ast.expr, rhs_ty: Type) -> ast.expr:
        """Helper function to check assignments with patterns."""
        raise InternalGuppyError("Unexpected assignment pattern")

    @_check_assign.register
    def _check_variable_assign(
        self, lhs: ast.Name, _rhs: ast.expr, rhs_ty: Type
    ) -> PlaceNode:
        x = lhs.id
        var = Variable(x, rhs_ty, lhs)
        self.ctx.locals[x] = var
        return with_loc(lhs, with_type(rhs_ty, PlaceNode(place=var)))

    @_check_assign.register
    def _check_field_assign(
        self, lhs: ast.Attribute, _rhs: ast.expr, rhs_ty: Type
    ) -> PlaceNode:
        # Unfortunately, the `attr` is just a string,  not an AST node, so we
        # have to compute its span by hand. This is fine since linebreaks are
        # not allowed in the identifier following the `.`
        span = to_span(lhs)
        value, attr = lhs.value, lhs.attr
        attr_span = Span(span.end.shift_left(len(attr)), span.end)
        value, struct_ty = self._synth_expr(value)
        if not isinstance(struct_ty, StructType) or attr not in struct_ty.field_dict:
            raise GuppyTypeError(AttributeNotFoundError(attr_span, struct_ty, attr))
        field = struct_ty.field_dict[attr]
        # TODO: In the future, we could infer some type args here
        if field.ty != rhs_ty:
            # TODO: Get hold of a span for the RHS and use a regular `TypeMismatchError`
            #  instead (maybe with a custom hint).
            raise GuppyTypeError(AssignFieldTypeMismatchError(attr_span, rhs_ty, field))
        if not isinstance(value, PlaceNode):
            # For now we complain if someone tries to assign to something that is not a
            # place, e.g. `f().a = 4`. This would only make sense if there is another
            # reference to the return value of `f`, otherwise the mutation cannot be
            # observed. We can start supporting this once we have proper reference
            # semantics.
            err = UnsupportedError(value, "Assigning to this expression", singular=True)
            err.add_sub_diagnostic(AssignNonPlaceHelp(None, field))
            raise GuppyError(err)
        if field.ty.copyable:
            raise GuppyError(
                UnsupportedError(
                    attr_span, "Mutation of classical fields", singular=True
                )
            )
        place = FieldAccess(value.place, struct_ty.field_dict[attr], lhs)
        place = check_place_assignable(place, self.ctx, lhs, "assignable")
        return with_loc(lhs, with_type(rhs_ty, PlaceNode(place=place)))

    @_check_assign.register
    def _check_subscript_assign(
        self, lhs: ast.Subscript, rhs: ast.expr, rhs_ty: Type
    ) -> PlaceNode:
        # Check subscript is array subscript.
        value, container_ty = self._synth_expr(lhs.value)
        if not is_array_type(container_ty):
            raise GuppyError(
                UnsupportedError(lhs, "Subscript assignments to non-arrays")
            )

        # Check array element type matches type of RHS.
        element_ty = get_element_type(container_ty)
        if element_ty != rhs_ty:
            raise GuppyTypeError(
                AssignSubscriptTypeMismatchError(to_span(lhs), rhs_ty, element_ty)
            )

        # As with field assignment, only allow place assignments for now.
        if not isinstance(value, PlaceNode):
            raise GuppyError(
                UnsupportedError(value, "Assigning to this expression", singular=True)
            )

        # Create a subscript place
        item_expr, item_ty = self._synth_expr(lhs.slice)
        item = Variable(next(tmp_vars), item_ty, item_expr)
        place = SubscriptAccess(value.place, item, rhs_ty, item_expr)

        # Calling `check_place_assignable` makes sure that `__setitem__` is implemented
        place = check_place_assignable(place, self.ctx, lhs, "assignable")
        return with_loc(lhs, with_type(rhs_ty, PlaceNode(place=place)))

    @_check_assign.register
    def _check_tuple_assign(
        self, lhs: ast.Tuple, rhs: ast.expr, rhs_ty: Type
    ) -> AnyUnpack:
        return self._check_unpack_assign(lhs, rhs, rhs_ty)

    @_check_assign.register
    def _check_list_assign(
        self, lhs: ast.List, rhs: ast.expr, rhs_ty: Type
    ) -> AnyUnpack:
        return self._check_unpack_assign(lhs, rhs, rhs_ty)

    def _check_unpack_assign(
        self, lhs: ast.Tuple | ast.List, rhs: ast.expr, rhs_ty: Type
    ) -> AnyUnpack:
        """Helper function to check unpacking assignments.

        These are the ones where the LHS is either a tuple or a list.
        """
        # Parse LHS into `left, *starred, right`
        pattern = parse_unpack_pattern(lhs)
        left, starred, right = pattern.left, pattern.starred, pattern.right
        # Check that the RHS has an appropriate type to be unpacked
        unpack, rhs_elts, rhs_tys = self._check_unpackable(rhs, rhs_ty, pattern)

        # Check that the numbers match up on the LHS and RHS
        num_lhs, num_rhs = len(right) + len(left), len(rhs_tys)
        err = WrongNumberOfUnpacksError(
            lhs, num_rhs, num_lhs, at_least=starred is not None
        )
        if num_lhs > num_rhs:
            # Build span that covers the unexpected elts on the LHS
            span = Span(to_span(lhs.elts[num_rhs]).start, to_span(lhs.elts[-1]).end)
            raise GuppyTypeError(replace(err, span=span))
        elif num_lhs < num_rhs and not starred:
            raise GuppyTypeError(err)

        # Recursively check any nested patterns on the left or right
        le, rs = len(left), len(rhs_elts) - len(right)  # left_end, right_start
        unpack.pattern.left = [
            self._check_assign(pat, elt, ty)
            for pat, elt, ty in zip(left, rhs_elts[:le], rhs_tys[:le], strict=True)
        ]
        unpack.pattern.right = [
            self._check_assign(pat, elt, ty)
            for pat, elt, ty in zip(right, rhs_elts[rs:], rhs_tys[rs:], strict=True)
        ]

        # Starred assignments are collected into an array
        if starred:
            starred_tys = rhs_tys[le:rs]
            assert all_equal(starred_tys)
            if starred_tys:
                starred_ty, *_ = starred_tys
            # Starred part could be empty. If it's an iterable unpack, we're still fine
            # since we know the yielded type
            elif isinstance(unpack, IterableUnpack):
                starred_ty = unpack.compr.elt_ty
            # For tuple unpacks, there is no way to infer a type for the empty starred
            # part
            else:
                unsolved = array_type(ExistentialTypeVar.fresh("T", True, True), 0)
                raise GuppyError(TypeInferenceError(starred, unsolved))
            array_ty = array_type(starred_ty, len(starred_tys))
            unpack.pattern.starred = self._check_assign(starred, rhs_elts[0], array_ty)

        return with_type(rhs_ty, with_loc(lhs, unpack))

    def _check_unpackable(
        self, expr: ast.expr, ty: Type, pattern: UnpackPattern
    ) -> tuple[AnyUnpack, list[ast.expr], Sequence[Type]]:
        """Checks that the given expression can be used in an unpacking assignment.

        This is the case for expressions with tuple types or ones that are iterable with
        a static size. Also checks that the expression is compatible with the given
        unpacking pattern.

        Returns an AST node capturing the unpacking operation together with expressions
        and types for all unpacked items. Emits a user error if the given expression is
        not unpackable.
        """
        left, starred, right = pattern.left, pattern.starred, pattern.right
        if isinstance(ty, TupleType):
            # Starred assignment of tuples is only allowed if all starred elements have
            # the same type
            if starred:
                starred_tys = (
                    ty.element_types[len(left) : -len(right)]
                    if right
                    else ty.element_types[len(left) :]
                )
                if not all_equal(starred_tys):
                    tuple_ty = TupleType(starred_tys)
                    raise GuppyError(StarredTupleUnpackError(starred, tuple_ty))
            tys = ty.element_types
            elts = expr.elts if isinstance(expr, ast.Tuple) else [expr] * len(tys)
            return TupleUnpack(pattern), elts, tys

        elif self.ctx.globals.get_instance_func(ty, "__iter__"):
            size = check_iter_unpack_has_static_size(expr, self.ctx)
            # Create a dummy variable and assign the expression to it. This helps us to
            # wire it up correctly during Hugr generation.
            var = self._check_assign(make_var(next(tmp_vars), expr), expr, ty)
            assert isinstance(var, PlaceNode)
            # We collect the whole RHS into an array. For this, we can reuse the
            # existing array comprehension logic.
            elt = make_var(next(tmp_vars), expr)
            gen = ast.comprehension(target=elt, iter=var, ifs=[], is_async=False)
            [gen], elt = desugar_comprehension([with_loc(expr, gen)], elt, expr)
            # Type check the comprehension
            [gen], elt, elt_ty = synthesize_comprehension(expr, [gen], elt, self.ctx)
            compr = DesugaredArrayComp(
                elt, gen, length=ConstValue(nat_type(), size), elt_ty=elt_ty
            )
            compr = with_type(array_type(elt_ty, size), compr)
            return IterableUnpack(pattern, compr, var), size * [elt], size * [elt_ty]

        # Otherwise, we can't unpack this expression
        raise GuppyError(UnpackableError(expr, ty))

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        if len(node.targets) > 1:
            # This is the case for assignments like `a = b = 1`
            raise GuppyError(UnsupportedError(node, "Multi assignments"))

        [target] = node.targets
        node.value, ty = self._synth_expr(node.value)
        node.targets = [self._check_assign(target, node.value, ty)]
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.stmt:
        if node.value is None:
            raise GuppyError(UnsupportedError(node, "Variable declarations"))
        ty = type_from_ast(node.annotation, self.ctx.globals, self.ctx.generic_params)
        node.value, subst = self._check_expr(node.value, ty)
        assert not ty.unsolved_vars  # `ty` must be closed!
        assert len(subst) == 0
        target = self._check_assign(node.target, node.value, ty)
        return with_loc(node, ast.Assign(targets=[target], value=node.value))

    def visit_AugAssign(self, node: ast.AugAssign) -> ast.stmt:
        bin_op = with_loc(
            node, ast.BinOp(left=node.target, op=node.op, right=node.value)
        )
        assign = with_loc(node, ast.Assign(targets=[node.target], value=bin_op))
        return self.visit_Assign(assign)

    def visit_Expr(self, node: ast.Expr) -> ast.stmt:
        # An expression statement where the return value is discarded
        node.value, _ = self._synth_expr(node.value)
        return node

    def visit_Return(self, node: ast.Return) -> ast.stmt:
        if not self.return_ty:
            raise InternalGuppyError("return_ty required to check return stmt!")

        if node.value is not None:
            node.value, subst = self._check_expr(
                node.value, self.return_ty, "return value"
            )
            assert len(subst) == 0  # `self.return_ty` is closed!
        elif not isinstance(self.return_ty, NoneType):
            raise GuppyTypeError(MissingReturnValueError(node, self.return_ty))
        return node

    def visit_NestedFunctionDef(self, node: NestedFunctionDef) -> ast.stmt:
        from guppylang.checker.func_checker import check_nested_func_def

        if not self.bb:
            raise InternalGuppyError("BB required to check nested function def!")

        func_def = check_nested_func_def(node, self.bb, self.ctx)
        self.ctx.locals[func_def.name] = Variable(func_def.name, func_def.ty, func_def)
        return func_def

    def visit_If(self, node: ast.If) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")

    def visit_While(self, node: ast.While) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")

    def visit_Break(self, node: ast.Break) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")

    def visit_Continue(self, node: ast.Continue) -> None:
        raise InternalGuppyError("Control-flow statement should not be present here.")


T = TypeVar("T")


def all_equal(xs: Iterable[T]) -> bool:
    """Checks if all elements yielded from an iterable are equal."""
    it = iter(xs)
    try:
        first = next(it)
    except StopIteration:
        return True
    return all(first == x for x in it)


def parse_unpack_pattern(lhs: ast.Tuple | ast.List) -> UnpackPattern:
    """Parses the LHS of an unpacking assignment like `a, *bs, c = ...` or
    `[a, *bs, c] = ...`."""
    # Split up LHS into `left, *starred, right` (the Python grammar ensures
    # that there is at most one starred expression)
    left = list(takewhile(lambda e: not isinstance(e, ast.Starred), lhs.elts))
    starred = (
        cast(ast.Starred, lhs.elts[len(left)]).value
        if len(left) < len(lhs.elts)
        else None
    )
    right = lhs.elts[len(left) + 1 :]
    assert isinstance(starred, ast.Name | None), "Python grammar"
    return UnpackPattern(left, starred, right)


def check_iter_unpack_has_static_size(expr: ast.expr, ctx: Context) -> int:
    """Helper function to check that an iterable expression is suitable to be unpacked
    in an assignment.

    This is the case if the iterator has a static, non-generic size.

    Returns the size of the iterator or emits a user error if the iterable is not
    suitable.
    """
    expr_synth = ExprSynthesizer(ctx)
    make_iter = with_loc(expr, MakeIter(expr, expr, unwrap_size_hint=False))
    make_iter, iter_ty = expr_synth.visit_MakeIter(make_iter)
    err = UnpackableError(expr, get_type(expr))
    if not is_sized_iter_type(iter_ty):
        err.add_sub_diagnostic(UnpackableError.NonStaticIter(None))
        raise GuppyError(err)
    match get_iter_size(iter_ty):
        case ConstValue(value=int(size)):
            return size
        case generic_size:
            err.add_sub_diagnostic(UnpackableError.GenericSize(None, generic_size))
            raise GuppyError(err)



from typing import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(f: "Callable[[], qubit @owned]") -> None: ...


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule


module = GuppyModule("test")

T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo() -> T:
    ...


@guppy.declare(module)
def bar(x: T, y: T) -> None:
    ...


@guppy(module)
def main() -> None:
    bar(foo(), 42)


module.compile()



import ast
import functools
from collections.abc import Sequence

import hugr.tys as ht
from hugr import Wire, ops
from hugr.build.dfg import DfBase

from guppylang.ast_util import AstVisitor, get_type
from guppylang.checker.core import Variable, contains_subscript
from guppylang.compiler.core import (
    CompilerBase,
    CompilerContext,
    DFContainer,
    return_var,
)
from guppylang.compiler.expr_compiler import ExprCompiler
from guppylang.error import InternalGuppyError
from guppylang.nodes import (
    CheckedNestedFunctionDef,
    IterableUnpack,
    PlaceNode,
    TupleUnpack,
)
from guppylang.std._internal.compiler.array import (
    array_discard_empty,
    array_new,
    array_pop,
)
from guppylang.std._internal.compiler.prelude import build_unwrap
from guppylang.tys.builtin import get_element_type
from guppylang.tys.const import ConstValue
from guppylang.tys.ty import TupleType, Type, type_to_row


class StmtCompiler(CompilerBase, AstVisitor[None]):
    """A compiler for Guppy statements to Hugr"""

    expr_compiler: ExprCompiler

    dfg: DFContainer

    def __init__(self, ctx: CompilerContext):
        super().__init__(ctx)
        self.expr_compiler = ExprCompiler(ctx)

    def compile_stmts(
        self,
        stmts: Sequence[ast.stmt],
        dfg: DFContainer,
    ) -> DFContainer:
        """Compiles a list of basic statements into a dataflow node.

        Note that the `dfg` is mutated in-place. After compilation, the DFG will also
        contain all variables that are assigned in the given list of statements.
        """
        self.dfg = dfg
        for s in stmts:
            self.visit(s)
        return self.dfg

    @property
    def builder(self) -> DfBase[ops.DfParentOp]:
        """The Hugr dataflow graph builder."""
        return self.dfg.builder

    @functools.singledispatchmethod
    def _assign(self, lhs: ast.expr, port: Wire) -> None:
        """Updates the local DFG with assignments."""
        raise InternalGuppyError("Invalid assign pattern in compiler")

    @_assign.register
    def _assign_place(self, lhs: PlaceNode, port: Wire) -> None:
        if subscript := contains_subscript(lhs.place):
            assert subscript.setitem_call is not None
            if subscript.item not in self.dfg:
                self.dfg[subscript.item] = self.expr_compiler.compile(
                    subscript.item_expr, self.dfg
                )
            # If the subscript is nested inside the place, e.g. `xs[i].y = ...`, we
            # first need to lookup `tmp = xs[i]`, assign `tmp.y = ...`, and then finally
            # set `xs[i] = tmp`
            if subscript != lhs.place:
                assert subscript.getitem_call is not None
                # Instead of `tmp` just use `xs[i]` as a "name", the dfg tracker doesn't
                # care about this
                self.dfg[subscript] = self.expr_compiler.compile(
                    subscript.getitem_call, self.dfg
                )
            # Assign to the name `xs[i].y`
            self.dfg[lhs.place] = port
            # Look up `xs[i]` again since it was mutated by the assignment above, then
            # compile a call to `__setitem__` to actually mutate
            self.dfg[subscript.setitem_call.value_var] = self.dfg[subscript]
            self.expr_compiler.visit(subscript.setitem_call.call)
        else:
            self.dfg[lhs.place] = port

    @_assign.register
    def _assign_tuple(self, lhs: TupleUnpack, port: Wire) -> None:
        """Handles assignment where the RHS is a tuple that should be unpacked."""
        # Unpack the RHS tuple
        left, starred, right = lhs.pattern.left, lhs.pattern.starred, lhs.pattern.right
        types = [ty.to_hugr() for ty in type_to_row(get_type(lhs))]
        unpack = self.builder.add_op(ops.UnpackTuple(types), port)
        ports = list(unpack)

        # Assign left and right
        for pat, wire in zip(left, ports[: len(left)], strict=True):
            self._assign(pat, wire)
        if right:
            for pat, wire in zip(right, ports[-len(right) :], strict=True):
                self._assign(pat, wire)

        # Starred assignments are collected into an array
        if starred:
            array_ty = get_type(starred)
            starred_ports = (
                ports[len(left) : -len(right)] if right else ports[len(left) :]
            )
            elt = get_element_type(array_ty).to_hugr()
            opts = [self.builder.add_op(ops.Some(elt), p) for p in starred_ports]
            array = self.builder.add_op(array_new(ht.Option(elt), len(opts)), *opts)
            self._assign(starred, array)

    @_assign.register
    def _assign_iterable(self, lhs: IterableUnpack, port: Wire) -> None:
        """Handles assignment where the RHS is an iterable that should be unpacked."""
        # Given an assignment pattern `left, *starred, right`, collect the RHS into an
        # array and pop from the left and right, leaving us with the starred array in
        # the middle
        assert isinstance(lhs.compr.length, ConstValue)
        length = lhs.compr.length.value
        assert isinstance(length, int)
        opt_elt_ty = ht.Option(lhs.compr.elt_ty.to_hugr())

        def pop(
            array: Wire, length: int, pats: list[ast.expr], from_left: bool
        ) -> tuple[Wire, int]:
            err = "Internal error: unpacking of iterable failed"
            num_pats = len(pats)
            # Pop the number of requested elements from the array
            elts = []
            for i in range(num_pats):
                res = self.builder.add_op(
                    array_pop(opt_elt_ty, length - i, from_left), array
                )
                [elt_opt, array] = build_unwrap(self.builder, res, err)
                [elt] = build_unwrap(self.builder, elt_opt, err)
                elts.append(elt)
            # Assign elements to the given patterns
            for pat, elt in zip(
                pats,
                # Assignments are evaluated from left to right, so we need to assign in
                # reverse order if we popped from the right
                elts if from_left else reversed(elts),
                strict=True,
            ):
                self._assign(pat, elt)
            return array, length - num_pats

        self.dfg[lhs.rhs_var.place] = port
        array = self.expr_compiler.visit_DesugaredArrayComp(lhs.compr)
        array, length = pop(array, length, lhs.pattern.left, True)
        array, length = pop(array, length, lhs.pattern.right, False)
        if lhs.pattern.starred:
            self._assign(lhs.pattern.starred, array)
        else:
            assert length == 0
            self.builder.add_op(array_discard_empty(opt_elt_ty), array)

    def visit_Assign(self, node: ast.Assign) -> None:
        [target] = node.targets
        port = self.expr_compiler.compile(node.value, self.dfg)
        self._assign(target, port)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        assert node.value is not None
        port = self.expr_compiler.compile(node.value, self.dfg)
        self._assign(node.target, port)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_Expr(self, node: ast.Expr) -> None:
        self.expr_compiler.compile_row(node.value, self.dfg)

    def visit_Return(self, node: ast.Return) -> None:
        # We turn returns into assignments of dummy variables, i.e. the statement
        # `return e0, e1, e2` is turned into `%ret0 = e0; %ret1 = e1; %ret2 = e2`.
        if node.value is not None:
            return_ty = get_type(node.value)
            port = self.expr_compiler.compile(node.value, self.dfg)

            row: list[tuple[Wire, Type]]
            if isinstance(return_ty, TupleType):
                types = [e.to_hugr() for e in return_ty.element_types]
                unpack = self.builder.add_op(ops.UnpackTuple(types), port)
                row = list(zip(unpack, return_ty.element_types, strict=True))
            else:
                row = [(port, return_ty)]

            for i, (wire, ty) in enumerate(row):
                var = Variable(return_var(i), ty, node.value)
                self.dfg[var] = wire

    def visit_CheckedNestedFunctionDef(self, node: CheckedNestedFunctionDef) -> None:
        from guppylang.compiler.func_compiler import compile_local_func_def

        var = Variable(node.name, node.ty, node)
        loaded_func = compile_local_func_def(node, self.dfg, self.ctx)
        self.dfg[var] = loaded_func



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy(module)
def test(q: qubit) -> qubit:
    r = q
    q = qubit()
    return r


module.compile()



import guppylang.std.quantum as quantum
from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

module = GuppyModule("test")
module.load_all(quantum)


@guppy.declare(module)
def bar(q: qubit) -> list[qubit]: ...


@guppy(module)
def foo(qs: list[qubit] @owned) -> list[qubit]:
    return [r for q in qs for r in bar(q)]


module.compile()



from typing import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule

from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


T = guppy.type_var("T", module=module)


@guppy.declare(module)
def foo(x: Callable[[T], None]) -> None:
    ...

@guppy.declare(module)
def h(q: qubit) -> None: ...


@guppy(module)
def main() -> None:
    foo(h)


module.compile()



from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar

T = TypeVar("T")


class ToHugr(ABC, Generic[T]):
    """Abstract base class for objects that have a Hugr representation."""

    @abstractmethod
    def to_hugr(self) -> T:
        """Computes the Hugr representation of the object."""


class Visitor(ABC):
    """Abstract base class for a type visitor that transforms types."""

    @abstractmethod
    def visit(self, arg: Any, /) -> bool:
        """This method is called for each visited type.

        Return `False` to continue the recursive descent.
        """


class Transformer(ABC):
    """Abstract base class for a type visitor that transforms types."""

    @abstractmethod
    def transform(self, arg: Any, /) -> Any | None:
        """Transforms the object.

        Return a transformed type or `None` to continue the recursive visit.
        """


class Visitable(ABC):
    """Abstract base class for objects that can be recursively visited."""

    @abstractmethod
    def visit(self, visitor: Visitor, /) -> None:
        """Accepts a visitor on the object.

        Implementors of this method should first invoke the visitor on the object
        itself. If the visitor doesn't handle the object, the visitor should be passed
        to all relevant members of the object.
        """


class Transformable(Visitable, ABC, Generic[T]):
    """Abstract base class for objects that can be recursively transformed."""

    @abstractmethod
    def transform(self, transformer: Transformer, /) -> T:
        """Accepts a transformer on the object.

        Implementors of this method should first invoke the transformer on the object
        itself. If the visitor doesn't handle the object, the visitor should be used to
        transform all relevant members of the object.
        """



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.struct(module)
class MyStruct:
    q: qubit

@guppy(module)
def test(q: qubit) -> MyStruct:
    return MyStruct(q)


module.compile()



# ruff: noqa: F821
# ^ Stop ruff complaining about not knowing what "tensor" is

from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule


def test_check_callable(validate):
    module = GuppyModule("module")

    @guppy(module)
    def bar(f: Callable[[int], bool]) -> Callable[[int], bool]:
        return f

    @guppy(module)
    def foo(f: Callable[[int], bool]) -> tuple[Callable[[int], bool]]:
        return (f,)

    @guppy(module)
    def is_42(x: int) -> bool:
        return x == 42

    @guppy(module)
    def baz(x: int) -> bool:
        return foo(is_42)(x)

    @guppy(module)
    def baz1() -> tuple[Callable[[int], bool]]:
        return foo(is_42)

    @guppy(module)
    def baz2(x: int) -> bool:
        return (foo,)(is_42)(x)

    validate(module.compile())


def test_call(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo() -> int:
        return 42

    @guppy(module)
    def bar() -> bool:
        return True

    @guppy(module)
    def baz_ho() -> tuple[Callable[[], int], Callable[[], bool]]:
        return (foo, bar)

    @guppy(module)
    def baz_ho_id() -> tuple[Callable[[], int], Callable[[], bool]]:
        return baz_ho()

    @guppy(module)
    def baz_ho_call() -> tuple[int, bool]:
        return baz_ho_id()()

    @guppy(module)
    def baz() -> tuple[int, bool]:
        return (foo, bar)()

    @guppy(module)
    def call_var() -> tuple[int, bool, int]:
        f = foo
        g = (foo, bar, f)
        return g()

    validate(module.compile())


def test_call_inplace(validate):
    module = GuppyModule("module")

    @guppy(module)
    def local(f: Callable[[int], bool], g: Callable[[bool], int]) -> tuple[bool, int]:
        return (f, g)(42, True)

    validate(module.compile())


def test_singleton(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int, y: int) -> tuple[int, int]:
        return y, x

    @guppy(module)
    def baz(x: int) -> tuple[int, int]:
        return (foo,)(x, x)

    validate(module.compile())


def test_call_back(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> int:
        return x

    @guppy(module)
    def bar(x: int) -> int:
        return x * x

    @guppy(module)
    def baz(x: int, y: int) -> tuple[int, int]:
        return (foo, bar)(x, y)

    @guppy(module)
    def call_var(x: int) -> tuple[int, int, int]:
        f = foo, baz
        return f(x, x, x)

    validate(module.compile())


def test_normal(validate):
    module = GuppyModule("module")

    @guppy(module)
    def glo(x: int) -> int:
        return x

    @guppy(module)
    def foo(x: int) -> int:
        return glo(x)

    validate(module.compile())


def test_higher_order(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int) -> bool:
        return x > 42

    @guppy(module)
    def bar(x: float) -> int:
        if x < 5.0:
            return 0
        else:
            return 1

    @guppy(module)
    def baz() -> tuple[Callable[[int], bool], Callable[[float], int]]:
        return foo, bar

    # For the future:
    #
    #     @guppy(module)
    #     def apply(f: Callable[[int, float],
    #               tuple[bool, int]],
    #               args: tuple[int, float]) -> tuple[bool, int]:
    #         return f(*args)
    #
    #     @guppy(module)
    #     def apply_call(args: tuple[int, float]) -> tuple[bool, int]:
    #         return apply(baz, args)

    validate(module.compile())


def test_nesting(validate):
    module = GuppyModule("module")

    @guppy(module)
    def foo(x: int, y: int) -> int:
        return x + y

    @guppy(module)
    def bar(x: int) -> int:
        return -x

    @guppy(module)
    def call(x: int) -> tuple[int, int, int]:
        f = bar, (bar, foo)
        return f(x, x, x, x)

    validate(module.compile())



"""Native arithmetic operations from the HUGR std, and compilers for non native ones."""

from collections.abc import Sequence

import hugr.std.int
from hugr import ops
from hugr import tys as ht
from hugr.std.int import int_t

from guppylang.std._internal.compiler.prelude import error_type
from guppylang.tys.ty import NumericType

INT_T = int_t(NumericType.INT_WIDTH)

# ------------------------------------------------------
# --------- std.arithmetic.int operations --------------
# ------------------------------------------------------


def _instantiate_int_op(
    name: str,
    int_width: int | Sequence[int],
    inp: list[ht.Type],
    out: list[ht.Type],
) -> ops.ExtOp:
    op_def = hugr.std.int.INT_OPS_EXTENSION.get_op(name)
    int_width = [int_width] if isinstance(int_width, int) else int_width
    return ops.ExtOp(
        op_def,
        ht.FunctionType(inp, out),
        [ht.BoundedNatArg(w) for w in int_width],
    )


def ieq(width: int) -> ops.ExtOp:
    """Returns a `std.arithmetic.int.ieq` operation."""
    return _instantiate_int_op("ieq", width, [int_t(width), int_t(width)], [ht.Bool])


def ine(width: int) -> ops.ExtOp:
    """Returns a `std.arithmetic.int.ine` operation."""
    return _instantiate_int_op("ine", width, [int_t(width), int_t(width)], [ht.Bool])


def iwiden_u(from_width: int, to_width: int) -> ops.ExtOp:
    """Returns an unsigned `std.arithmetic.int.widen_u` operation."""
    return _instantiate_int_op(
        "iwiden_u", [from_width, to_width], [int_t(from_width)], [int_t(to_width)]
    )


def iwiden_s(from_width: int, to_width: int) -> ops.ExtOp:
    """Returns a signed `std.arithmetic.int.widen_s` operation."""
    return _instantiate_int_op(
        "iwiden_s", [from_width, to_width], [int_t(from_width)], [int_t(to_width)]
    )


def inarrow_u(from_width: int, to_width: int) -> ops.ExtOp:
    """Returns an unsigned `std.arithmetic.int.narrow_u` operation."""
    return _instantiate_int_op(
        "inarrow_u",
        [from_width, to_width],
        [int_t(from_width)],
        [ht.Either([error_type()], [int_t(to_width)])],
    )


def inarrow_s(from_width: int, to_width: int) -> ops.ExtOp:
    """Returns a signed `std.arithmetic.int.narrow_s` operation."""
    return _instantiate_int_op(
        "inarrow_s",
        [from_width, to_width],
        [int_t(from_width)],
        [ht.Either([error_type()], [int_t(to_width)])],
    )


# ------------------------------------------------------
# --------- std.arithmetic.conversions ops -------------
# ------------------------------------------------------


def _instantiate_convert_op(
    name: str,
    inp: list[ht.Type],
    out: list[ht.Type],
    args: list[ht.TypeArg] | None = None,
) -> ops.ExtOp:
    op_def = hugr.std.int.CONVERSIONS_EXTENSION.get_op(name)
    return ops.ExtOp(op_def, ht.FunctionType(inp, out), args or [])


def convert_ifromusize() -> ops.ExtOp:
    """Returns a `std.arithmetic.conversions.ifromusize` operation."""
    return _instantiate_convert_op("ifromusize", [ht.USize()], [INT_T])


def convert_itousize() -> ops.ExtOp:
    """Returns a `std.arithmetic.conversions.itousize` operation."""
    return _instantiate_convert_op("itousize", [INT_T], [ht.USize()])


def convert_ifrombool() -> ops.ExtOp:
    """Returns a `std.arithmetic.conversions.ifrombool` operation."""
    return _instantiate_convert_op("ifrombool", [ht.Bool], [int_t(0)])


def convert_itobool() -> ops.ExtOp:
    """Returns a `std.arithmetic.conversions.itobool` operation."""
    return _instantiate_convert_op("itobool", [int_t(0)], [ht.Bool])



from __future__ import annotations

import inspect
import sys
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from pathlib import Path
from types import FrameType, ModuleType
from typing import TYPE_CHECKING, Any

from hugr.build.function import Module
from hugr.package import ModulePointer, Package

import guppylang.compiler.hugr_extension
from guppylang import decorator
from guppylang.checker.core import Globals, PyScope
from guppylang.compiler.core import CompilerContext
from guppylang.definition.common import (
    CheckableDef,
    CheckedDef,
    DefId,
    Definition,
    ParsableDef,
    RawDef,
)
from guppylang.definition.declaration import RawFunctionDecl
from guppylang.definition.function import RawFunctionDef
from guppylang.definition.module import ModuleDef
from guppylang.definition.parameter import ParamDef
from guppylang.definition.pytket_circuits import RawPytketDef
from guppylang.definition.struct import CheckedStructDef
from guppylang.definition.ty import TypeDef
from guppylang.error import pretty_errors
from guppylang.experimental import enable_experimental_features

if TYPE_CHECKING:
    from hugr import Hugr, ops

    from guppylang.span import SourceMap

PyClass = type
PyFunc = Callable[..., Any]
PyFuncDefOrDecl = tuple[bool, PyFunc]


class GuppyModule:
    """A Guppy module that may contain function and type definitions."""

    name: str

    # Whether the module has already been checked
    _checked: bool

    # If the hugr has already been compiled, keeps a reference that can be returned
    # from `compile`.
    _compiled: _CompiledModule | None

    # Map of raw definitions in this module
    _raw_defs: dict[DefId, RawDef]
    _raw_type_defs: dict[DefId, RawDef]

    # Map of checked definitions in this module. These are populated by calling `check`
    # on this module
    _checked_defs: dict[DefId, CheckedDef]

    # Globals from imported modules
    _imported_globals: Globals
    # Checked definitions from imported modules. Also includes transitively imported
    # definitions.
    _imported_checked_defs: dict[DefId, CheckedDef]

    # Globals for functions and types defined in this module. Only gets populated during
    # compilation
    _globals: Globals

    # When `_instance_buffer` is not `None`, then all registered functions will be
    # buffered in this list. They only get properly registered, once
    # `_register_buffered_instance_funcs` is called. This way, we can associate
    _instance_func_buffer: dict[str, RawDef] | None

    # Storage for source code that has been read by the compiler
    _sources: SourceMap

    def __init__(self, name: str, import_builtins: bool = True):
        self.name = name
        self._globals = Globals({}, {}, {}, {})
        self._imported_globals = Globals.default()
        self._imported_checked_defs = {}
        self._checked = False
        self._compiled = None
        self._instance_func_buffer = None
        self._raw_defs = {}
        self._raw_type_defs = {}
        self._checked_defs = {}

        from guppylang.decorator import guppy

        self._sources = guppy._sources

        # Import builtin module
        if import_builtins:
            import guppylang.std.builtins as builtins

            # Std lib is allowed to use experimental features
            with enable_experimental_features():
                self.load_all(builtins)

    def load(
        self,
        *args: Definition | GuppyModule | ModuleType,
        **kwargs: Definition | GuppyModule | ModuleType,
    ) -> None:
        """Imports another Guppy module or selected definitions from a module.

        Keyword args may be used to specify alias names for the imports.
        """
        # Note that we shouldn't evaluate imports during a sphinx build since the @guppy
        # decorator is mocked in that case
        if sphinx_running():
            return

        modules: set[GuppyModule] = set()
        defs: dict[DefId, CheckedDef] = {}
        names: dict[str, DefId] = {}

        # Collect imports in reverse since we'll use it as a stack to push and pop from
        imports: list[tuple[str, Definition | GuppyModule | ModuleType]] = [
            *reversed(kwargs.items()),
            *(("", arg) for arg in reversed(args)),
        ]
        while imports:
            alias, imp = imports.pop()
            if isinstance(imp, Definition):
                module = imp.id.module
                assert module is not None
                module.check()
                names[alias or imp.name] = imp.id
                modules.add(module)
            elif isinstance(imp, GuppyModule):
                imp.check()
                def_id = DefId.fresh(imp)
                name = alias or imp.name
                defn = ModuleDef(def_id, name, None, imp._globals)
                defs[def_id] = defn
                names[name] = def_id
                modules.add(imp)
            elif isinstance(imp, ModuleType):
                mod = find_guppy_module_in_py_module(imp)
                imports.append((alias, mod))
            else:
                msg = f"Only Guppy definitions or modules can be imported. Got `{imp}`"
                raise TypeError(msg)

        # Also include any impls that are defined by the imported modules
        impls: dict[DefId, dict[str, DefId]] = {}
        for module in modules:
            # We need to include everything defined in the module, including stuff that
            # is not directly imported, in order to lower everything into a single Hugr
            defs |= module._imported_checked_defs
            defs |= module._checked_defs
            # We also need to include any impls that are transitively imported
            all_globals = module._imported_globals | module._globals
            for def_id in all_globals.impls:
                impls.setdefault(def_id, {})
                impls[def_id] |= all_globals.impls[def_id]
        self._imported_globals |= Globals(dict(defs), names, impls, {})
        self._imported_checked_defs |= defs

        # We also need to include transitively imported checked definitions so we can
        # lower everything into one Hugr at the same time.
        for module in modules:
            self._imported_checked_defs |= module._imported_checked_defs

    def load_all(self, mod: GuppyModule | ModuleType) -> None:
        """Imports all public members of a module."""
        # Note that we shouldn't evaluate imports during a sphinx build since the @guppy
        # decorator is mocked in that case
        if sphinx_running():
            return

        if isinstance(mod, GuppyModule):
            mod.check()
            self.load(
                *(
                    defn
                    for defn in mod._globals.defs.values()
                    if not defn.name.startswith("_")
                )
            )
        elif isinstance(mod, ModuleType):
            self.load_all(find_guppy_module_in_py_module(mod))
        else:
            msg = f"Only Guppy definitions or modules can be imported. Got `{mod}`"
            raise TypeError(msg)

    def register_def(self, defn: RawDef, instance: TypeDef | None = None) -> None:
        """Registers a definition with this module.

        If the name of the definition is already defined, the new definition
        replaces the old.

        Optionally, the definition can be marked as an instance method by passing the
        corresponding instance type definition.
        """
        self._checked = False
        self._compiled = None
        if self._instance_func_buffer is not None and not isinstance(defn, TypeDef):
            self._instance_func_buffer[defn.name] = defn
        else:
            # If this overrides an already defined name, we need to purge the old
            # definition to avoid checking it later
            if self.contains(defn.name):
                self.unregister(self._globals[defn.name])
            if isinstance(defn, TypeDef | ParamDef):
                self._raw_type_defs[defn.id] = defn
            else:
                self._raw_defs[defn.id] = defn
            if instance is not None:
                self._globals.impls.setdefault(instance.id, {})
                self._globals.impls[instance.id][defn.name] = defn.id
            else:
                self._globals.names[defn.name] = defn.id
            self._globals.defs[defn.id] = defn

    def register_func_def(
        self, f: PyFunc, instance: TypeDef | None = None
    ) -> RawFunctionDef:
        """Registers a Python function definition as belonging to this Guppy module."""
        defn = RawFunctionDef(DefId.fresh(self), f.__name__, None, f, get_py_scope(f))
        self.register_def(defn, instance)
        return defn

    def register_func_decl(
        self, f: PyFunc, instance: TypeDef | None = None
    ) -> RawFunctionDecl:
        """Registers a Python function declaration as belonging to this Guppy module."""
        decl = RawFunctionDecl(DefId.fresh(self), f.__name__, None, f, get_py_scope(f))
        self.register_def(decl, instance)
        return decl

    def register_pytket_func(
        self, f: PyFunc, input_value: Any, instance: TypeDef | None = None
    ) -> RawPytketDef:
        """Registers a pytket circuit function as belonging to this Guppy module."""
        func = RawPytketDef(
            DefId.fresh(self), f.__name__, None, f, get_py_scope(f), input_value
        )
        self.register_def(func, instance)
        return func

    def _register_buffered_instance_funcs(self, instance: TypeDef) -> None:
        assert self._instance_func_buffer is not None
        buffer = self._instance_func_buffer
        self._instance_func_buffer = None
        for defn in buffer.values():
            self.register_def(defn, instance)

    def unregister(self, defn: Definition) -> None:
        """Removes a definition from this module.

        Also removes all methods when unregistering a type.
        """
        self._checked = False
        self._compiled = None
        self._raw_defs.pop(defn.id, None)
        self._raw_type_defs.pop(defn.id, None)
        self._globals.defs.pop(defn.id, None)
        self._globals.names.pop(defn.name, None)
        if impls := self._globals.impls.pop(defn.id, None):
            for impl in impls.values():
                self.unregister(self._globals[impl])

    @property
    def checked(self) -> bool:
        return self._checked

    @property
    def compiled(self) -> bool:
        return self._compiled is not None

    def _check_defs(
        self, raw_defs: Mapping[DefId, RawDef], globals: Globals
    ) -> dict[DefId, CheckedDef]:
        """Helper method to parse and check raw definitions."""
        raw_globals = globals | Globals(dict(raw_defs), {}, {}, {})
        parsed = {
            def_id: defn.parse(raw_globals, self._sources)
            if isinstance(defn, ParsableDef)
            else defn
            for def_id, defn in raw_defs.items()
        }
        parsed_globals = globals | Globals(dict(parsed), {}, {}, {})
        return {
            def_id: (
                defn.check(parsed_globals) if isinstance(defn, CheckableDef) else defn
            )
            for def_id, defn in parsed.items()
        }

    @pretty_errors
    def check(self) -> None:
        """Type-checks the module."""
        if self.checked:
            return

        # Type definitions need to be checked first so that we can use them when parsing
        # function signatures etc.
        type_defs = self._check_defs(
            self._raw_type_defs, self._imported_globals | self._globals
        )
        self._globals.defs.update(type_defs)

        # Collect auto-generated methods
        generated: dict[DefId, RawDef] = {}
        for defn in type_defs.values():
            if isinstance(defn, CheckedStructDef):
                self._globals.impls.setdefault(defn.id, {})
                for method_def in defn.generated_methods():
                    generated[method_def.id] = method_def
                    self._globals.impls[defn.id][method_def.name] = method_def.id
        self._globals.defs.update(generated)

        # Now, we can check all other definitions
        other_defs = self._check_defs(
            self._raw_defs | generated, self._imported_globals | self._globals
        )
        self._globals.defs.update(other_defs)
        self._checked_defs = type_defs | other_defs
        self._checked = True

    def compile_hugr(self) -> Hugr[ops.Module]:
        """Compiles the module and returns the final Hugr."""
        # This function does not use the `pretty_errors` decorator since it is
        # is wrapping around `compile_package` which does use it already.
        return self.compile().module

    @pretty_errors
    def compile(self) -> ModulePointer:
        """Compiles the module and returns the final Hugr package.

        The package contains the single Hugr graph as well as the required
        extensions definitions.
        """
        if self._compiled is not None:
            return self._compiled.module

        self.check()
        checked_defs = self._imported_checked_defs | self._checked_defs

        # Prepare Hugr for this module
        graph = Module()
        graph.metadata["name"] = self.name

        # Lower definitions to Hugr
        ctx = CompilerContext(
            checked_defs, graph, self._imported_globals | self._globals
        )
        for defn in self._checked_defs.values():
            ctx.compile(defn)

        # TODO: Currently we just include a hardcoded list of extensions. We should
        # compute this dynamically from the imported dependencies instead.
        #
        # The hugr prelude and std_extensions are implicit.
        from guppylang.std._internal.compiler.quantum import TKET2_EXTENSIONS

        extensions = [*TKET2_EXTENSIONS, guppylang.compiler.hugr_extension.EXTENSION]

        mod_ptr = ModulePointer(Package(modules=[graph.hugr], extensions=extensions), 0)
        self._compiled = _CompiledModule(ctx, mod_ptr)
        return self._compiled.module

    def contains(self, name: str) -> bool:
        """Returns 'True' if the module contains an object with the given name."""
        return name in self._globals.names


@dataclass(frozen=True)
class _CompiledModule:
    """Cached compiled module and definitions"""

    context: CompilerContext
    module: ModulePointer


def get_py_scope(f: PyFunc) -> PyScope:
    """Returns a mapping of all variables captured by a Python function together with
    the `f_locals` and `f_globals` of the frame that called this function.

    Note that this function only works in CPython. On other platforms, an empty
    dictionary is returned.

    Relies on inspecting the `__globals__` and `__closure__` attributes of the function.
    See https://docs.python.org/3/reference/datamodel.html#special-read-only-attributes
    """
    # Get variables from the calling frame
    frame = get_calling_frame()
    frame_vars = frame.f_globals | frame.f_locals if frame else {}

    if sys.implementation.name != "cpython":
        return frame_vars

    if inspect.ismethod(f):
        f = f.__func__
    code = f.__code__

    nonlocals: PyScope = {}
    if f.__closure__ is not None:
        for var, cell in zip(code.co_freevars, f.__closure__, strict=True):
            try:
                value = cell.cell_contents
            except ValueError:
                # The call to `cell_contents` will fail if `var` is a recursive
                # reference to the decorated function
                continue
            nonlocals[var] = value

    return frame_vars | nonlocals | f.__globals__.copy()


def find_guppy_module_in_py_module(module: ModuleType) -> GuppyModule:
    """Helper function to search the `__dict__` of a Python module for an instance of
     `GuppyModule`.

    Raises a user-error if no unique module can be found.
    """
    mods = [val for val in module.__dict__.values() if isinstance(val, GuppyModule)]
    # Also include implicit modules
    from guppylang.decorator import ModuleIdentifier, guppy

    if hasattr(module, "__file__") and module.__file__:
        module_id = ModuleIdentifier(Path(module.__file__), module.__name__, module)
        if module_id in guppy.registered_modules():
            mods.append(guppy.get_module(module_id))

    if not mods:
        msg = f"No Guppy modules found in `{module.__name__}`"
        raise ValueError(msg)
    if len(mods) > 1:
        msg = (
            f"Python module `{module.__name__}` contains multiple Guppy modules. "
            "Cannot decide which one to import."
        )
        raise ValueError(msg)
    return mods[0]


def get_calling_frame() -> FrameType | None:
    """Finds the first frame that called this function outside the compiler modules."""
    frame = inspect.currentframe()
    while frame:
        module = inspect.getmodule(frame)
        if module is None:
            break
        if module.__file__ != __file__ and module != decorator:
            return frame
        frame = frame.f_back
    return None


def sphinx_running() -> bool:
    """Checks if this module was imported during a sphinx build."""
    # This is the most general solution available at the moment.
    # See: https://github.com/sphinx-doc/sphinx/issues/9805
    try:
        import sphinx  # type: ignore[import-untyped, import-not-found, unused-ignore]

        return hasattr(sphinx, "application")
    except ImportError:
        return False



from collections.abc import Callable

from tests.util import compile_guppy


@compile_guppy
def foo(x: int) -> int:
    def bar(f: Callable[[int], bool]) -> bool:
        return f(42)

    return bar(foo)



"""Type checking code for control-flow graphs

Operates on CFGs produced by the `CFGBuilder`. Produces a `CheckedCFG` consisting of
`CheckedBB`s with inferred type signatures.
"""

import ast
import collections
from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import ClassVar, Generic, TypeVar

from guppylang.ast_util import line_col
from guppylang.cfg.bb import BB
from guppylang.cfg.cfg import CFG, BaseCFG
from guppylang.checker.core import Context, Globals, Locals, Place, V, Variable
from guppylang.checker.expr_checker import ExprSynthesizer, to_bool
from guppylang.checker.stmt_checker import StmtChecker
from guppylang.definition.value import ValueDef
from guppylang.diagnostic import Error, Note
from guppylang.error import GuppyError
from guppylang.tys.param import Parameter
from guppylang.tys.ty import InputFlags, Type

Row = Sequence[V]


@dataclass(frozen=True)
class Signature(Generic[V]):
    """The signature of a basic block.

    Stores the input/output variables with their types. Generic over the representation
    of program variables.
    """

    input_row: Row[V]
    output_rows: Sequence[Row[V]]  # One for each successor

    dummy_output_rows: Sequence[Row[V]] = field(default_factory=list)

    @staticmethod
    def empty() -> "Signature[V]":
        return Signature([], [], [])


@dataclass(eq=False)  # Disable equality to recover hash from `object`
class CheckedBB(BB, Generic[V]):
    """Basic block annotated with an input and output type signature.

    The signature is generic over the representation of program variables.
    """

    sig: Signature[V] = field(default_factory=Signature.empty)


class CheckedCFG(BaseCFG[CheckedBB[V]], Generic[V]):
    input_tys: list[Type]
    output_ty: Type

    def __init__(self, input_tys: list[Type], output_ty: Type) -> None:
        super().__init__([])
        self.input_tys = input_tys
        self.output_ty = output_ty


def check_cfg(
    cfg: CFG,
    inputs: Row[Variable],
    return_ty: Type,
    generic_params: dict[str, Parameter],
    func_name: str,
    globals: Globals,
) -> CheckedCFG[Place]:
    """Type checks a control-flow graph.

    Annotates the basic blocks with input and output type signatures and removes
    unreachable blocks. Note that the inputs/outputs are annotated in the form of
    *places* rather than just variables.
    """
    # First, we need to run program analysis
    ass_before = {v.name for v in inputs}
    inout_vars = [v for v in inputs if InputFlags.Inout in v.flags]
    cfg.analyze(ass_before, ass_before, [v.name for v in inout_vars])

    # We start by compiling the entry BB
    checked_cfg: CheckedCFG[Variable] = CheckedCFG([v.ty for v in inputs], return_ty)
    checked_cfg.entry_bb = check_bb(
        cfg.entry_bb, checked_cfg, inputs, return_ty, generic_params, globals
    )
    compiled = {cfg.entry_bb: checked_cfg.entry_bb}

    # Visit all control-flow edges in BFS order. We can't just do a normal loop over
    # all BBs since the input types for a BB are computed by checking a predecessor.
    # We do BFS instead of DFS to get a better error ordering.
    queue = collections.deque(
        (checked_cfg.entry_bb, i, succ)
        # We enumerate the successor starting from the back, so we start with the `True`
        # branch. This way, we find errors in a more natural order
        for i, succ in reverse_enumerate(
            cfg.entry_bb.successors + cfg.entry_bb.dummy_successors
        )
    )
    while len(queue) > 0:
        pred, num_output, bb = queue.popleft()
        pred_outputs = [*pred.sig.output_rows, *pred.sig.dummy_output_rows]
        input_row = [
            Variable(v.name, v.ty, v.defined_at, v.flags)
            for v in pred_outputs[num_output]
        ]

        if bb in compiled:
            # If the BB was already compiled, we just have to check that the signatures
            # match.
            check_rows_match(input_row, compiled[bb].sig.input_row, bb, globals)
        else:
            # Otherwise, check the BB and enqueue its successors
            checked_bb = check_bb(
                bb, checked_cfg, input_row, return_ty, generic_params, globals
            )
            queue += [
                # We enumerate the successor starting from the back, so we start with
                # the `True` branch. This way, we find errors in a more natural order
                (checked_bb, i, succ)
                for i, succ in reverse_enumerate(bb.successors)
            ]
            compiled[bb] = checked_bb

        # Link up BBs in the checked CFG, excluding the unreachable ones
        if bb.reachable:
            compiled[bb].predecessors.append(pred)
            pred.successors[num_output] = compiled[bb]

    # The exit BB might be unreachable. In that case it won't be visited above and we
    # have to handle it here
    if cfg.exit_bb not in compiled:
        assert not cfg.exit_bb.reachable
        compiled[cfg.exit_bb] = CheckedBB(
            cfg.exit_bb.idx, checked_cfg, reachable=False, sig=Signature(inout_vars, [])
        )

    required_bbs = [bb for bb in cfg.bbs if bb.reachable or bb.is_exit]
    checked_cfg.bbs = [compiled[bb] for bb in required_bbs]
    checked_cfg.exit_bb = compiled[cfg.exit_bb]
    checked_cfg.live_before = {compiled[bb]: cfg.live_before[bb] for bb in required_bbs}
    checked_cfg.ass_before = {compiled[bb]: cfg.ass_before[bb] for bb in required_bbs}
    checked_cfg.maybe_ass_before = {
        compiled[bb]: cfg.maybe_ass_before[bb] for bb in required_bbs
    }

    # Finally, run the linearity check
    from guppylang.checker.linearity_checker import check_cfg_linearity

    linearity_checked_cfg = check_cfg_linearity(checked_cfg, func_name, globals)
    return linearity_checked_cfg


@dataclass(frozen=True)
class VarNotDefinedError(Error):
    title: ClassVar[str] = "Variable not defined"
    span_label: ClassVar[str] = "`{var}` is not defined"
    var: str


@dataclass(frozen=True)
class VarMaybeNotDefinedError(Error):
    title: ClassVar[str] = "Variable not defined"
    var: str

    @dataclass(frozen=True)
    class BadBranch(Note):
        span_label: ClassVar[str] = "... if this expression is `{truth_value}`"
        var: str
        truth_value: bool

    @property
    def rendered_span_label(self) -> str:
        s = f"`{self.var}` might be undefined"
        if self.children:
            s += " ..."
        return s


@dataclass(frozen=True)
class BranchTypeError(Error):
    title: ClassVar[str] = "Different types"
    span_label: ClassVar[str] = "{ident} may refer to different types"
    ident: str

    @dataclass(frozen=True)
    class TypeHint(Note):
        span_label: ClassVar[str] = "This is of type `{ty}`"
        ty: Type

    @dataclass(frozen=True)
    class GlobalHint(Note):
        message: ClassVar[str] = (
            "{ident} may be shadowing a global {defn.description} definition of type "
            "`{defn.ty}` on some branches"
        )
        defn: ValueDef


@dataclass(frozen=True)
class GlobalShadowError(Error):
    title: ClassVar[str] = "Global variable conditionally shadowed"
    span_label: ClassVar[str] = "{ident} may be shadowing a global variable"
    ident: str


def check_bb(
    bb: BB,
    checked_cfg: CheckedCFG[Variable],
    inputs: Row[Variable],
    return_ty: Type,
    generic_params: dict[str, Parameter],
    globals: Globals,
) -> CheckedBB[Variable]:
    cfg = bb.containing_cfg

    # For the entry BB we have to separately check that all used variables are
    # defined. For all other BBs, this will be checked when compiling a predecessor.
    if bb == cfg.entry_bb:
        assert len(bb.predecessors) == 0
        for x, use in bb.vars.used.items():
            if x not in cfg.ass_before[bb] and x not in globals:
                raise GuppyError(VarNotDefinedError(use, x))

    # Check the basic block
    ctx = Context(globals, Locals({v.name: v for v in inputs}), generic_params)
    checked_stmts = StmtChecker(ctx, bb, return_ty).check_stmts(bb.statements)

    # If we branch, we also have to check the branch predicate
    if len(bb.successors) > 1:
        assert bb.branch_pred is not None
        bb.branch_pred, ty = ExprSynthesizer(ctx).synthesize(bb.branch_pred)
        bb.branch_pred, _ = to_bool(bb.branch_pred, ty, ctx)

    for succ in bb.successors + bb.dummy_successors:
        for x, use_bb in cfg.live_before[succ].items():
            # Check that the variables requested by the successor are defined
            if x not in ctx.locals and x not in ctx.globals:
                # If the variable is defined on *some* paths, we can give a more
                # informative error message
                if x in cfg.maybe_ass_before[use_bb]:
                    err = VarMaybeNotDefinedError(use_bb.vars.used[x], x)
                    if bad_branch := diagnose_maybe_undefined(use_bb, x, cfg):
                        branch_expr, truth_value = bad_branch
                        note = VarMaybeNotDefinedError.BadBranch(
                            branch_expr, x, truth_value
                        )
                        err.add_sub_diagnostic(note)
                    raise GuppyError(err)
                raise GuppyError(VarNotDefinedError(use_bb.vars.used[x], x))

    # Finally, we need to compute the signature of the basic block
    outputs = [
        [ctx.locals[x] for x in cfg.live_before[succ] if x in ctx.locals]
        for succ in bb.successors
    ]
    dummy_outputs = [
        [ctx.locals[x] for x in cfg.live_before[succ] if x in ctx.locals]
        for succ in bb.dummy_successors
    ]

    # Also prepare the successor list so we can fill it in later
    checked_bb = CheckedBB(
        bb.idx,
        checked_cfg,
        checked_stmts,
        reachable=bb.reachable,
        sig=Signature(inputs, outputs, dummy_outputs),
    )
    checked_bb.successors = [None] * len(bb.successors)  # type: ignore[list-item]
    checked_bb.branch_pred = bb.branch_pred
    return checked_bb


def check_rows_match(
    row1: Row[Variable], row2: Row[Variable], bb: BB, globals: Globals
) -> None:
    """Checks that the types of two rows match up.

    Otherwise, an error is thrown, alerting the user that a variable has different
    types on different control-flow paths.
    """
    map1, map2 = {v.name: v for v in row1}, {v.name: v for v in row2}
    for x in map1.keys() | map2.keys():
        # If block signature lengths don't match but no undefined error was thrown, some
        # variables may be shadowing global variables.
        v1 = map1.get(x) or globals[x]
        assert isinstance(v1, Variable | ValueDef)
        v2 = map2.get(x) or globals[x]
        assert isinstance(v2, Variable | ValueDef)
        if v1.ty != v2.ty:
            # In the error message, we want to mention the variable that was first
            # defined at the start.
            if (
                v1.defined_at
                and v2.defined_at
                and line_col(v2.defined_at) < line_col(v1.defined_at)
            ):
                v1, v2 = v2, v1
            # We shouldn't mention temporary variables (starting with `%`)
            # in error messages:
            ident = "Expression" if v1.name.startswith("%") else f"Variable `{v1.name}`"
            use = bb.containing_cfg.live_before[bb][v1.name].vars.used[v1.name]
            err = BranchTypeError(use, ident)
            # We don't add a location to the type hint for the global variable,
            # since it could lead to cross-file diagnostics (which are not
            # supported) or refer to long function definitions.
            sub1 = (
                BranchTypeError.TypeHint(v1.defined_at, v1.ty)
                if isinstance(v1, Variable)
                else BranchTypeError.GlobalHint(None, v1)
            )
            sub2 = (
                BranchTypeError.TypeHint(v2.defined_at, v2.ty)
                if isinstance(v2, Variable)
                else BranchTypeError.GlobalHint(None, v2)
            )
            err.add_sub_diagnostic(sub1)
            err.add_sub_diagnostic(sub2)
            raise GuppyError(err)
        else:
            # TODO: Remove once https://github.com/CQCL/guppylang/issues/827 is done.
            # If either is a global variable, don't allow shadowing even if types match.
            if not (isinstance(v1, Variable) and isinstance(v2, Variable)):
                local_var = v1 if isinstance(v1, Variable) else v2
                ident = (
                    "Expression"
                    if local_var.name.startswith("%")
                    else f"Variable `{local_var.name}`"
                )
                glob_err = GlobalShadowError(local_var.defined_at, ident)
                raise GuppyError(glob_err)


def diagnose_maybe_undefined(
    bb: BB, x: str, cfg: BaseCFG[BB]
) -> tuple[ast.expr, bool] | None:
    """Given a BB and a variable `x`, tries to find a branch where one of the successors
    leads to an assignment of `x` while the other one does not.

    Returns the branch condition and a flag whether the value being `True` leads to the
    undefined path. Returns `None` if no such branch can be found.
    """
    assert x in cfg.maybe_ass_before[bb]
    # Find all BBs that can reach this BB and which ones of those assign `x`
    ancestors = list(cfg.ancestors(bb))
    assigns = [anc for anc in ancestors if x in anc.vars.assigned]
    # Compute which ancestors can possibly reach an assignment
    reaches_assignment = set(cfg.ancestors(*assigns))
    # Try to find a branching BB where one of paths can reach an assignment, while the
    # other one cannot
    for anc in ancestors:
        match anc.successors:
            case [true_succ, false_succ]:
                assert anc.branch_pred is not None
                true_reaches_assignment = true_succ in reaches_assignment
                false_reaches_assignment = false_succ in reaches_assignment
                if true_reaches_assignment != false_reaches_assignment:
                    return anc.branch_pred, true_reaches_assignment
    return None


T = TypeVar("T")


def reverse_enumerate(xs: list[T]) -> Iterator[tuple[int, T]]:
    """Enumerates a list in reverse order.

    Equivalent to `reversed(list(enumerate(data)))` without creating an intermediate
    list.
    """
    for i in range(len(xs) - 1, -1, -1):
        yield i, xs[i]



from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(x: qubit) -> qubit: ...


@guppy(module)
def test() -> Callable[[qubit @owned], qubit]:
    return foo


module.compile()



from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

module = GuppyModule("test")
module.load(qubit)


@guppy.declare(module)
def foo(q: qubit) -> None: ...


@guppy.struct(module)
class MyImmutableContainer:
    q: qubit

    @guppy.declare(module)
    def __getitem__(self: "MyImmutableContainer", idx: int) -> qubit: ...


@guppy(module)
def test(c: MyImmutableContainer) -> MyImmutableContainer:
    foo(c[0])
    return c


module.compile()



import ast
from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar

from hugr import Wire, ops
from hugr import tys as ht
from hugr.build.dfg import DfBase

from guppylang.ast_util import AstNode, get_type, has_empty_body, with_loc, with_type
from guppylang.checker.core import Context, Globals
from guppylang.checker.expr_checker import check_call, synthesize_call
from guppylang.checker.func_checker import check_signature
from guppylang.compiler.core import CompilerContext, DFContainer
from guppylang.definition.common import ParsableDef
from guppylang.definition.value import CallReturnWires, CompiledCallableDef
from guppylang.diagnostic import Error, Help
from guppylang.error import GuppyError, InternalGuppyError
from guppylang.nodes import GlobalCall
from guppylang.span import SourceMap
from guppylang.tys.subst import Inst, Subst
from guppylang.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    Type,
    type_to_row,
)

if TYPE_CHECKING:
    from guppylang.definition.function import PyFunc


@dataclass(frozen=True)
class BodyNotEmptyError(Error):
    title: ClassVar[str] = "Unexpected function body"
    span_label: ClassVar[str] = "Body of custom function `{name}` must be empty"
    name: str


@dataclass(frozen=True)
class NoSignatureError(Error):
    title: ClassVar[str] = "Type signature missing"
    span_label: ClassVar[str] = "Custom function `{name}` requires a type signature"
    name: str

    @dataclass(frozen=True)
    class Suggestion(Help):
        message: ClassVar[str] = (
            "Annotate the type signature of `{name}` or disallow the use of `{name}` "
            "as a higher-order value: `@guppy.custom(..., higher_order_value=False)`"
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(NoSignatureError.Suggestion(None))


@dataclass(frozen=True)
class NotHigherOrderError(Error):
    title: ClassVar[str] = "Not higher-order"
    span_label: ClassVar[str] = (
        "Function `{name}` may not be used as a higher-order value"
    )
    name: str


@dataclass(frozen=True)
class RawCustomFunctionDef(ParsableDef):
    """A raw custom function definition provided by the user.

    Custom functions provide their own checking and compilation logic using a
    `CustomCallChecker` and a `CustomCallCompiler`.

    The raw definition stores exactly what the user has written (i.e. the AST together
    with the provided checker and compiler), without inspecting the signature.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
        call_checker: The custom call checker.
        call_compiler: The custom call compiler.
        higher_order_value: Whether the function may be used as a higher-order value.
    """

    python_func: "PyFunc"
    call_checker: "CustomCallChecker"
    call_compiler: "CustomInoutCallCompiler"

    # Whether the function may be used as a higher-order value. This is only possible
    # if a static type for the function is provided.
    higher_order_value: bool

    description: str = field(default="function", init=False)

    def parse(self, globals: "Globals", sources: SourceMap) -> "CustomFunctionDef":
        """Parses and checks the user-provided signature of the custom function.

        The signature is optional if custom type checking logic is provided by the user.
        However, note that the signature annotation is required, if we want to use the
        function as a higher-order value.

        If no signature is provided, we fill in the dummy signature `() -> ()`. This
        type will never be inspected, since we rely on the provided custom checking
        code. The only information we need to access is that it's a function type and
        that there are no unsolved existential vars.
        """
        from guppylang.definition.function import parse_py_func

        func_ast, docstring = parse_py_func(self.python_func, sources)
        if not has_empty_body(func_ast):
            raise GuppyError(BodyNotEmptyError(func_ast.body[0], self.name))
        sig = self._get_signature(func_ast, globals)
        ty = sig or FunctionType([], NoneType())
        return CustomFunctionDef(
            self.id,
            self.name,
            func_ast,
            ty,
            self.call_checker,
            self.call_compiler,
            self.higher_order_value,
            sig is not None,
        )

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
        function_ty: ht.FunctionType,
    ) -> Sequence[Wire]:
        """Compiles a call to the function."""
        # Note: We have _compiled_ globals rather than `Globals` here,
        # so we cannot use `self._get_signature()` to build a `CustomFunctionDef`.
        self.call_compiler._setup(
            type_args,
            dfg,
            ctx,
            node,
            function_ty,
            None,
        )
        return self.call_compiler.compile_with_inouts(args).regular_returns

    def _get_signature(
        self, node: ast.FunctionDef, globals: Globals
    ) -> FunctionType | None:
        """Returns the type of the function, if known.

        Type annotations are needed if we rely on the default call checker or
        want to allow the usage of the function as a higher-order value.

        Some function types like python's `int()` cannot be expressed in the Guppy
        type system, so we return `None` here and rely on the specialized compiler
        to handle the call.
        """
        requires_type_annotation = (
            isinstance(self.call_checker, DefaultCallChecker) or self.higher_order_value
        )
        has_type_annotation = node.returns or any(
            arg.annotation for arg in node.args.args
        )

        if requires_type_annotation and not has_type_annotation:
            raise GuppyError(NoSignatureError(node, self.name))

        if requires_type_annotation:
            return check_signature(node, globals)
        else:
            return None


@dataclass(frozen=True)
class CustomFunctionDef(CompiledCallableDef):
    """A custom function with parsed and checked signature.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
        ty: The type of the function. This may be a dummy value if `has_signature` is
            false.
        call_checker: The custom call checker.
        call_compiler: The custom call compiler.
        higher_order_value: Whether the function may be used as a higher-order value.
        has_signature: Whether the function has a declared signature.
    """

    defined_at: AstNode
    ty: FunctionType
    call_checker: "CustomCallChecker"
    call_compiler: "CustomInoutCallCompiler"
    higher_order_value: bool
    has_signature: bool

    description: str = field(default="function", init=False)

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type.

        This is done by invoking the provided `CustomCallChecker`.
        """
        self.call_checker._setup(ctx, node, self)
        new_node, subst = self.call_checker.check(args, ty)
        return with_type(ty, with_loc(node, new_node)), subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call.

        This is done by invoking the provided `CustomCallChecker`.
        """
        self.call_checker._setup(ctx, node, self)
        new_node, ty = self.call_checker.synthesize(args)
        return with_type(ty, with_loc(node, new_node)), ty

    def load_with_args(
        self,
        type_args: Inst,
        dfg: "DFContainer",
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the custom function as a value into a local dataflow graph.

        This will place a `FunctionDef` node in the local DFG, and load with a
        `LoadFunc` node. This operation will fail if the function is not allowed
        to be used as a higher-order value.
        """
        # TODO: This should be raised during checking, not compilation!
        if not self.higher_order_value:
            raise GuppyError(NotHigherOrderError(node, self.name))
        assert len(self.ty.params) == len(type_args)

        # We create a `FunctionDef` that takes some inputs, compiles a call to the
        # function, and returns the results. If the function signature is polymorphic,
        # we explicitly monomorphise here and invoke the call compiler with the
        # inferred type args.
        #
        # TODO: Reuse compiled instances with the same type args?
        # TODO: Why do we need to monomorphise here? Why not wait for `load_function`?
        # See https://github.com/CQCL/guppylang/issues/393 for both issues.
        fun_ty = self.ty.instantiate(type_args)
        input_types = [inp.ty.to_hugr() for inp in fun_ty.inputs]
        output_types = [fun_ty.output.to_hugr()]
        func = dfg.builder.define_function(
            self.name, input_types, output_types, type_params=[]
        )

        func_dfg = DFContainer(func, dfg.locals.copy())
        args: list[Wire] = list(func.inputs())
        returns = self.compile_call(args, type_args, func_dfg, ctx, node)

        func.set_outputs(*returns.regular_returns, *returns.inout_returns)

        # Finally, load the function into the local DFG. We already monomorphised, so we
        # don't need to give it type arguments.
        return dfg.builder.load_function(func)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: "DFContainer",
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        if self.has_signature:
            concrete_ty = self.ty.instantiate(type_args)
        else:
            assert isinstance(node, GlobalCall)
            concrete_ty = FunctionType(
                [FuncInput(get_type(arg), InputFlags.NoFlags) for arg in node.args],
                get_type(node),
            )
        hugr_ty = concrete_ty.to_hugr()

        self.call_compiler._setup(type_args, dfg, ctx, node, hugr_ty, self)
        return self.call_compiler.compile_with_inouts(args)


class CustomCallChecker(ABC):
    """Abstract base class for custom function call type checkers."""

    ctx: Context
    node: AstNode
    func: CustomFunctionDef

    def _setup(self, ctx: Context, node: AstNode, func: CustomFunctionDef) -> None:
        self.ctx = ctx
        self.node = node
        self.func = func

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        """Checks the return value against a given type.

        Returns a (possibly) transformed and annotated AST node for the call.
        """
        from guppylang.checker.expr_checker import check_type_against

        expr, res_ty = self.synthesize(args)
        expr, subst, _ = check_type_against(res_ty, ty, expr, self.ctx)
        return expr, subst

    @abstractmethod
    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        """Synthesizes a type for the return value of a call.

        Also returns a (possibly) transformed and annotated argument list.
        """


class CustomInoutCallCompiler(ABC):
    """Abstract base class for custom function call compilers with borrowed args.

    Args:
        builder: The function builder where the function should be defined.
        type_args: The type arguments for the function.
        globals: The compiled globals.
        node: The AST node where the function is defined.
        ty: The type of the function, if known.
    """

    dfg: DFContainer
    type_args: Inst
    ctx: CompilerContext
    node: AstNode
    ty: ht.FunctionType
    func: CustomFunctionDef | None

    def _setup(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
        hugr_ty: ht.FunctionType,
        func: CustomFunctionDef | None,
    ) -> None:
        self.type_args = type_args
        self.dfg = dfg
        self.ctx = ctx
        self.node = node
        self.ty = hugr_ty
        self.func = func

    @abstractmethod
    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        """Compiles a custom function call.

        Returns the outputs of the call together with any borrowed arguments that are
        passed through the function.
        """

    @property
    def builder(self) -> DfBase[ops.DfParentOp]:
        """The hugr dataflow builder."""
        return self.dfg.builder


class CustomCallCompiler(CustomInoutCallCompiler, ABC):
    """Abstract base class for custom function call compilers with only owned args."""

    @abstractmethod
    def compile(self, args: list[Wire]) -> list[Wire]:
        """Compiles a custom function call and returns the resulting ports.

        Use the provided `self.builder` to add nodes to the Hugr graph.
        """

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        return CallReturnWires(self.compile(args), inout_returns=[])


class DefaultCallChecker(CustomCallChecker):
    """Checks function calls by comparing to a type signature."""

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.func.ty, args, ty, self.node, self.ctx)
        return GlobalCall(def_id=self.func.id, args=args, type_args=inst), subst

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.func.ty, args, self.node, self.ctx)
        return GlobalCall(def_id=self.func.id, args=args, type_args=inst), ty


class NotImplementedCallCompiler(CustomCallCompiler):
    """Call compiler for custom functions that are already lowered during checking.

    For example, the custom checker could replace the call with a series of calls to
    other functions. In that case, the original function will no longer be present and
    thus doesn't need to be compiled.
    """

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Function should have been removed during checking")


class OpCompiler(CustomInoutCallCompiler):
    """Call compiler for functions that are directly implemented via Hugr ops.

    args:
        op: A function that takes an instantiation of the type arguments as well as
            the monomorphic function type, and returns a concrete HUGR op.
    """

    op: Callable[[ht.FunctionType, Inst], ops.DataflowOp]

    def __init__(self, op: Callable[[ht.FunctionType, Inst], ops.DataflowOp]) -> None:
        self.op = op

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        op = self.op(self.ty, self.type_args)
        node = self.builder.add_op(op, *args)
        num_returns = (
            len(type_to_row(self.func.ty.output)) if self.func else len(self.ty.output)
        )
        return CallReturnWires(
            regular_returns=list(node[:num_returns]),
            inout_returns=list(node[num_returns:]),
        )


class NoopCompiler(CustomCallCompiler):
    """Call compiler for functions that are noops."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        return args


class CopyInoutCompiler(CustomInoutCallCompiler):
    """Call compiler for functions that are noops but only want to borrow arguments."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        return CallReturnWires(regular_returns=args, inout_returns=args)



from typing import no_type_check

from guppylang.decorator import guppy
from guppylang.module import GuppyModule
from guppylang.std import angles
from guppylang.std._internal.compiler.quantum import (
    QSYSTEM_EXTENSION,
    InoutMeasureCompiler,
)
from guppylang.std._internal.util import quantum_op
from guppylang.std.angles import angle
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit

qsystem = GuppyModule("qsystem")

qsystem.load(qubit)
qsystem.load_all(angles)


@guppy(qsystem)
@no_type_check
def phased_x(q: qubit, angle1: angle, angle2: angle) -> None:
    f1 = float(angle1)
    f2 = float(angle2)
    _phased_x(q, f1, f2)


@guppy.hugr_op(quantum_op("ZZMax", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def zz_max(q1: qubit, q2: qubit) -> None: ...


@guppy(qsystem)
@no_type_check
def zz_phase(q1: qubit, q2: qubit, angle: angle) -> None:
    f = float(angle)
    _zz_phase(q1, q2, f)


@guppy(qsystem)
@no_type_check
def rz(q: qubit, angle: angle) -> None:
    f1 = float(angle)
    _rz(q, f1)


@guppy.hugr_op(quantum_op("Measure", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def measure(q: qubit @ owned) -> bool: ...


@guppy.custom(InoutMeasureCompiler("MeasureReset", QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def measure_and_reset(q: qubit) -> bool:
    """MeasureReset operation from the qsystem extension."""


@guppy.hugr_op(quantum_op("Reset", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def reset(q: qubit) -> None: ...


# TODO
# @guppy.hugr_op(quantum_op("TryQAlloc", ext=QSYSTEM_EXTENSION), module=qsystem)
# @no_type_check
# def _try_qalloc() -> Option[qubit]:
#     ..


@guppy.hugr_op(quantum_op("QFree", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def qfree(q: qubit @ owned) -> None: ...


# ------------------------------------------------------
# --------- Internal definitions -----------------------
# ------------------------------------------------------


@guppy.hugr_op(quantum_op("PhasedX", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def _phased_x(q: qubit, angle1: float, angle2: float) -> None:
    """PhasedX operation from the qsystem extension.

    See `guppylang.std.qsystem.phased_x` for a public definition that
    accepts angle parameters.
    """


@guppy.hugr_op(quantum_op("ZZPhase", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def _zz_phase(q1: qubit, q2: qubit, angle: float) -> None:
    """ZZPhase operation from the qsystem extension.

    See `guppylang.std.qsystem.phased_x` for a public definition that
    accepts angle parameters.
    """


@guppy.hugr_op(quantum_op("Rz", ext=QSYSTEM_EXTENSION), module=qsystem)
@no_type_check
def _rz(q: qubit, angle: float) -> None:
    """Rz operation from the qsystem extension.

    See `guppylang.std.qsystem.rz` for a public definition that
    accepts angle parameters.
    """
